"""
    Definition of base `SchemaView` class.
"""

import re
import json
import pathlib


class SchemaView:
    # pylint: disable=too-few-public-methods

    """
        Django schema view. Base class for `DjangoSchemaView`.
    """

    def __init__(self, schema, with_graphiql=True):
        self.schema = schema
        self.with_graphiql = with_graphiql
        if with_graphiql:
            graphiql_page_path = pathlib.Path(__file__).parent / 'static/graphiql.html'
            with open(graphiql_page_path, 'rt', encoding='utf-8') as graphiql_page_file:
                self.graphiql_page = graphiql_page_file.read()

    def _must_serve_graphiql(self, headers):
        if not self.with_graphiql:
            return False
        # taken from https://github.com/graphql-python/graphene-django/blob/
        # e1a7d1983314174c91ede1ebbfe35a9009cf6268/graphene_django/views.py#L33
        def qualify(content_type):
            parts = content_type.split(';', 1)
            if len(parts) == 2:
                match = re.match(r'(^|;)q=(0(\.\d{,3})?|1(\.0{,3})?)(;|$)', parts[1])
                if match:
                    return parts[0].strip(), float(match.group(2))
            return parts[0].strip(), 1
        raw_content_types = headers.get('Accept', '*/*').split(',')
        qualified_content_types = map(qualify, raw_content_types)
        accepted_content_types = [
            content_type[0]
            for content_type in sorted(
                qualified_content_types, key=lambda content_type: content_type[1], reverse=True)
        ]
        # taken from https://github.com/graphql-python/graphene-django/blob/
        # e1a7d1983314174c91ede1ebbfe35a9009cf6268/graphene_django/views.py#L351
        html_priority = (
            len(accepted_content_types) - accepted_content_types.index('text/html')
            if 'text/html' in accepted_content_types
            else 0
        )
        json_priority = (
            len(accepted_content_types) - accepted_content_types.index('application/json')
            if 'application/json' in accepted_content_types
            else 0
        )
        return html_priority > json_priority

    # pylint: disable=too-many-return-statements
    def compute_response(self, method, headers, body, query, authenticated_user):
        """
            Compute response to be served by HTTP server.
        """
        # data extraction
        if method == 'POST':
            # extract data from JSON payload
            try:
                data = json.loads(body)
            except json.decoder.JSONDecodeError as error:
                return {'errors': [{'message':
                    f'HTTP request body is not valid JSON: {error}',
                }]}, 400
            if not isinstance(data, dict):
                return {'errors': [{'message':
                    'HTTP request body should be formatted as a JSON object',
                }]}, 400
        elif method == 'GET':
            if self._must_serve_graphiql(headers):
                return self.graphiql_page
            try:
                data = {
                    'query': query.get('query'),
                    'variables': json.loads(query.get('variables', 'null')),
                    'operationName': query.get('operationName'),
                }
            except json.JSONDecodeError:
                return {'errors': [{'message':
                    f'Parameter `variables` is not valid JSON: {error}',
                }]}, 400
        else:
            return {'errors': [{'message':
                f'Method {method} not allowed, only GET and POST are supported',
            }]}, 405
        # extract & validate string query
        query = data.get('query')
        if not isinstance(query, str):
            return {'errors': [{'message':
                'Required parameter "query" should be a string',
            }]}, 400
        # extract & validate variables mapping
        variables = data.get('variables') or {}
        if not isinstance(variables, dict):
            return {'errors': [{'message':
                'Optional parameter "variables" should be an mapping',
            }]}, 400
        # extract & validate string query
        operation_name = data.get('operationName', None)
        if operation_name is not None and not isinstance(operation_name, str):
            return {'errors': [{'message':
                'Optional parameter "operationName" should be a string',
            }]}, 400
        # compute & return result
        result = self.schema.execute(
            query = query,
            variables = variables,
            operation_name = operation_name,
            authenticated_user = authenticated_user,
            serializable_output = True,
        )
        return result, 200



"""
    Useful tools for introspection, used internally.
"""

import inspect


def get_subclasses(cls, exclude=None):
    """
        Get subclasses of class, excluding itself.
    """
    assert inspect.isclass(cls)
    if exclude is None:
        exclude = set()
    exclude.add(cls)
    for subclass in cls.__subclasses__():
        if subclass not in exclude:
            yield subclass
            exclude.add(subclass)
            yield from get_subclasses(subclass, exclude)


def get_method_arguments(method, exclude=None):
    """
        Return the method's argument, associated with a boolean describing whether
        or not the given argument is required.
    """
    signature = inspect.getfullargspec(method)
    args_count = len(signature.args) if signature.args else 0
    default_args_count = len(signature.defaults) if signature.defaults else 0
    result = {
        arg: args_count - i > default_args_count
        for i, arg in enumerate(signature.args)
    }
    if exclude:
        for arg in exclude:
            result.pop(arg)
    return result

def get_public_class_attributes(cls):
    """
        Retrieve the attribute names of a class that do not have a leading underscore.
    """
    return {
        attribute: getattr(cls, attribute)
        for attribute in dir(cls)
        if attribute[0] != '_'
    }

def validate_class_attributes_against_method_arguments(cls, method,
        excluded_arguments=None, message=''):
    """
        Check if the public attributes of a class can be used as arguments for a method
    """
    attributes = get_public_class_attributes(cls)
    arguments = get_method_arguments(method, excluded_arguments)
    required_arguments = set(argument for argument, required in arguments.items() if required)
    for required_argument in required_arguments:
        if required_argument not in attributes:
            raise ValueError(f'Attribute `{required_argument}` should be present on '
                f'class `{cls}`{message}')
    for attribute in attributes:
        if attribute not in arguments:
            raise ValueError(f'Invalid attribute `{attribute}` for class `{cls}`{message}; '
                'consider making it private with a leading underscore')


def is_instance_or_subclass(thing, klass):
    """
        Check if a thing is an instance or subclass of a given class (or set of classes)
    """
    return isinstance(thing, klass) or (inspect.isclass(thing) and issubclass(thing, klass))



"""
    Abstract classes for later subclassing and exposition
"""


class Exposed: # pylint: disable=too-few-public-methods
    """
        Base class for `ExposedQuery`, `ExposedMutation` and `ExposedModel`
    """

class ExposedQuery(Exposed): # pylint: disable=too-few-public-methods
    """
        Abstract base class for exposing a query
    """

class ExposedMutation(Exposed): # pylint: disable=too-few-public-methods
    """
        Abstract base class for exposing a mutation
    """

class ExposedModel(Exposed): # pylint: disable=too-few-public-methods
    """
        Abstract base class for exposing a model
    """

class CustomField(Exposed): # pylint: disable=too-few-public-methods
    """
        Abstract base class for exposing a custom field for a model
    """



"""
    Definition of `PeeweeModelManager` class.
"""

from ._manager import ModelManager


# pylint:disable=W0223 # Methods are abstract in class 'ModelManager' but not overridden

class PeeweeModelManager(ModelManager):
    """
        ModelManager class for Peewee ORM.
    """

    # remove Pylint disabling above class declaration when overridding methods



"""
    This module imports GraphQL types from graphql-core, but also defines some
    additional custom types.
"""

import datetime
import decimal
import json
from typing import Any
from dateutil.parser.isoparser import DEFAULT_ISOPARSER

import graphql.type
# Had to disable pylint below, because "No name '...' in module '...'"
from graphql.language.ast import ValueNode, StringValueNode # pylint: disable=no-name-in-module
from graphql.language.printer import print_ast

# native scalar types

Boolean = graphql.type.GraphQLBoolean
Int = graphql.type.GraphQLInt
Float = graphql.type.GraphQLFloat
String = graphql.type.GraphQLString

# native scalar wrappers

Union = graphql.type.GraphQLUnionType
List = graphql.type.GraphQLList
NonNull = graphql.type.GraphQLNonNull

# decimal type

def serialize_decimal(output_value: decimal.Decimal) -> str:
    """ Serializes an internal value to include in a response. """
    return str(output_value)

def parse_decimal_value(input_value: Any) -> decimal.Decimal:
    """ Parses an externally provided value to use as an input. """
    try:
        return decimal.Decimal(input_value)
    except Exception as error:
        raise ValueError(
            f'Cannot parse Decimal from: {repr(input_value)}, got: {error}') from error

def parse_decimal_literal(value_node: ValueNode, _variables: Any = None) -> decimal.Decimal:
    """ Parses an externally provided AST value to use as an input. """
    if not isinstance(value_node, StringValueNode):
        raise ValueError(
            "Decimal should be represented as a string in input: " + print_ast(value_node),
            value_node,
        )
    return parse_decimal_value(value_node.value)

Decimal = graphql.type.GraphQLScalarType(
    name = 'Decimal',
    description = 'A decimal (fixed-point)',
    serialize = serialize_decimal,
    parse_value = parse_decimal_value,
    parse_literal = parse_decimal_literal,
)

# datetime type

def serialize_datetime(output_value: datetime.datetime) -> str:
    """ Serializes an internal value to include in a response. """
    return output_value.isoformat()

def parse_datetime_value(input_value: Any) -> datetime.datetime:
    """ Parses an externally provided value to use as an input. """
    try:
        return DEFAULT_ISOPARSER.isoparse(input_value)
    except Exception as error:
        raise ValueError(
            f'Cannot parse DateTime from: {repr(input_value)}, got: {error}') from error

def parse_datetime_literal(value_node: ValueNode, _variables: Any = None) -> datetime.datetime:
    """ Parses an externally provided AST value to use as an input. """
    if not isinstance(value_node, StringValueNode):
        raise ValueError(
            "DateTime should be represented as a string in input: " + print_ast(value_node),
            value_node,
        )
    return parse_datetime_value(value_node.value)

DateTime = graphql.type.GraphQLScalarType(
    name = 'DateTime',
    description = 'A datetime element, serialized in standard "YYYY-MM-DDThh:ii::ss.mmmmmm" format',
    serialize = serialize_datetime,
    parse_value = parse_datetime_value,
    parse_literal = parse_datetime_literal,
)

# date type

def serialize_date(output_value: datetime.date) -> str:
    """ Serializes an internal value to include in a response. """
    return output_value.isoformat()

def parse_date_value(input_value: Any) -> datetime.date:
    """ Parses an externally provided value to use as an input. """
    try:
        return DEFAULT_ISOPARSER.parse_isodate(input_value)
    except Exception as error:
        raise ValueError(f'Cannot parse Date from: {repr(input_value)}, got: {error}') from error

def parse_date_literal(value_node: ValueNode, _variables: Any = None) -> datetime.date:
    """ Parses an externally provided AST value to use as an input. """
    if not isinstance(value_node, StringValueNode):
        raise ValueError(
            "Date should be represented as a string in input: " + print_ast(value_node),
            value_node,
        )
    return parse_date_value(value_node.value)

Date = graphql.type.GraphQLScalarType(
    name = 'Date',
    description = 'A date element, serialized in standard "YYYY-MM-DD" format',
    serialize = serialize_date,
    parse_value = parse_date_value,
    parse_literal = parse_date_literal,
)

# time type

def serialize_time(output_value: datetime.time) -> str:
    """ Serializes an internal value to include in a response. """
    return output_value.isoformat()

def parse_time_value(input_value: Any) -> datetime.time:
    """ Parses an externally provided value to use as an input. """
    try:
        return DEFAULT_ISOPARSER.parse_isotime(input_value)
    except Exception as error:
        raise ValueError(f'Cannot parse Time from: {repr(input_value)}, got: {error}') from error

def parse_time_literal(value_node: ValueNode, _variables: Any = None) -> datetime.time:
    """ Parses an externally provided AST value to use as an input. """
    if not isinstance(value_node, StringValueNode):
        raise ValueError(
            "Time should be represented as a string in input: " + print_ast(value_node),
            value_node,
        )
    return parse_time_value(value_node.value)

Time = graphql.type.GraphQLScalarType(
    name = 'Time',
    description = 'A time element, serialized in standard "hh:ii::ss.mmmmmm" format',
    serialize = serialize_time,
    parse_value = parse_time_value,
    parse_literal = parse_time_literal,
)

# JSON type

def serialize_jsonstring(output_value: datetime.time) -> str:
    """ Serializes an internal value to include in a response. """
    return json.dumps(output_value)

def parse_jsonstring_value(input_value: Any) -> datetime.time:
    """ Parses an externally provided value to use as an input. """
    try:
        return json.loads(input_value)
    except Exception as error:
        raise ValueError(
            f'Cannot parse JSONString from: {repr(input_value)}, got: {error}') from error

def parse_jsonstring_literal(value_node: ValueNode, _variables: Any = None) -> datetime.time:
    """ Parses an externally provided AST value to use as an input. """
    if not isinstance(value_node, StringValueNode):
        raise ValueError(
            "JSONString should be represented as a string in input: " + print_ast(value_node),
            value_node,
        )
    return parse_jsonstring_value(value_node.value)

JSONString = graphql.type.GraphQLScalarType(
    name = 'JSONString',
    description = 'JSON in a string',
    serialize = serialize_jsonstring,
    parse_value = parse_jsonstring_value,
    parse_literal = parse_jsonstring_literal,
)



"""
    Definition of ContextValue class.
"""


class ContextValue:

    # pylint: disable=too-few-public-methods

    """
        `ContextValue` objects are passed as `context_value` parameter to `graphql_sync()`
        method when performing GraphQL queries.
    """

    def __init__(self, authenticated_user):
        self.authenticated_user = authenticated_user



"""
    Basic operations on ORM models are defined here.
"""


import enum


@enum.unique
class Operation(enum.Enum):
    """
        Basic operations that can be performed on an ORM model.
    """

    CREATE = 'create'
    READ = 'read'
    UPDATE = 'update'
    DELETE = 'delete'
    FILTER = 'filter'



"""
    Module for custom JSON serialization.
"""

import json
import datetime


class JSONEncoder(json.JSONEncoder):

    """
        Overrides `json.JSONEncoder`, with the benefit of being able to serialize
        date, time and datetime instances.
    """

    def default(self, o):
        """
            Overrides `json.JSONEncoder.default`, with the benefit of being able to serialize
            date, time and datetime instances.
        """
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        return super().default(o)


dumps = JSONEncoder().encode



"""
    easy_graphql_server is an interface to easily expose a database in GraphQL via ORM models.
"""

from .types import JSONString, Required, Model
from .schema import Schema
from .operations import Operation
from .exceptions import UnauthenticatedError, NotFoundError, ForbiddenError, \
    ValidationError, DuplicateError, IntegrityError
from .exposition import ExposedModel, ExposedQuery, ExposedMutation, CustomField

CREATE = Operation.CREATE
READ = Operation.READ
UPDATE = Operation.UPDATE
DELETE = Operation.DELETE



"""
    `Schema` class is defined here. This is the root class that will be used as
    a central point to the easy_graphql_server API.
"""

import re
import json
import inspect
from collections import defaultdict
import logging

from graphql import GraphQLSchema, GraphQLField, GraphQLObjectType
from graphql.type.validate import validate_schema # pylint: disable=no-name-in-module,import-error
from graphql.utilities import get_introspection_query # pylint: disable=no-name-in-module,import-error
from graphql.graphql import graphql_sync
from graphql.language.ast import FieldNode, InlineFragmentNode, FragmentSpreadNode

from . import exceptions, exposition, introspection
from .conversion import to_graphql_type, to_graphql_argument
from .model_config import ModelConfig
from .casing import Casing
from .context import ContextValue


class Schema:

    """
        Base class for easy_graphql_server API.

        Mutation methods can be exposed with `Schema.expose_mutation()`.

        Query methods can be exposed with `Schema.expose_query()`.

        ORM model can be exposed with `Schema.expose_model()`.

        Any subclass of `ExposedModel`, `ExposedQuery` or `ExposedMutation` can
        be exposed via `Schema.expose()`.

        Schema can be validated using `Schema.check()`.
    """

    _requests_logger = logging.getLogger('easy_graphql_server.requests')
    _processing_logger = logging.getLogger('easy_graphql_server.processing')

    # public methods

    def __init__(self, debug=False, casing=Casing.SNAKE, restrict_models_queried_fields=False,
        models_max_depth=None, models_limit=-1, models_allowed_lookups=None, models_disallowed_lookups=None):
        self.methods = defaultdict(dict)
        self.subclasses = []
        self.dirty = True
        self.graphql_schema = None
        self.models_configs = []
        # options
        self.case_manager = casing.value
        self.debug = debug
        self.restrict_models_queried_fields = restrict_models_queried_fields
        self.models_max_depth = models_max_depth
        self.models_limit = models_limit
        self.models_allowed_lookups = models_allowed_lookups
        self.models_disallowed_lookups = models_disallowed_lookups
        # abstract parent classes
        class Exposed(exposition.Exposed):
            # pylint: disable=too-few-public-methods,missing-class-docstring
            pass
        self.Exposed = Exposed # pylint: disable=invalid-name
        class ExposedModel(exposition.ExposedModel, self.Exposed):
            # pylint: disable=too-few-public-methods,missing-class-docstring
            pass
        self.ExposedModel = ExposedModel # pylint: disable=invalid-name
        class ExposedQuery(exposition.ExposedQuery, self.Exposed):
            # pylint: disable=too-few-public-methods,missing-class-docstring
            pass
        self.ExposedQuery = ExposedQuery # pylint: disable=invalid-name
        class ExposedMutation(exposition.ExposedMutation, self.Exposed):
            # pylint: disable=too-few-public-methods,missing-class-docstring
            pass
        self.ExposedMutation = ExposedMutation # pylint: disable=invalid-name
        self.orm_model_manager_classes = []

    def get_documentation(self, with_descriptions=False):
        """
            Return GraphQL schema description in JSON format.
        """
        result = self.execute(
            get_introspection_query(descriptions=with_descriptions, directive_is_repeatable=True))
        return result.data

    def expose(self, cls):
        """
            Expose a subclass from `exposition.ExposedQuery`, `exposition.ExposedMutation`
            or `exposition.ExposedModel`
        """
        if not inspect.isclass(cls):
            raise ValueError('Parameter to `expose()` method should be a class, '
                f'`{cls}` was given instead.')
        self._expose_from_subclass(cls)

    def expose_query(self, **options):
        """
            Expose a query method in the schema.

            See `_expose_method` for more info about options.
        """
        self._expose_method(type_='query', **options)

    def expose_mutation(self, **options):
        """
            Expose a mutation method in the schema.

            See `_expose_method` for more info about options.
        """
        self._expose_method(type_='mutation', **options)

    def expose_model(self, orm_model, **options):
        """
            Expose an ORM model in the schema.

            See `ModelConfig` class constructor for more info about options.
        """
        self.dirty = True
        if 'restrict_queried_fields' not in options:
            options['restrict_queried_fields'] = self.restrict_models_queried_fields
        for option_name in ('max_depth', 'limit', 'allowed_lookups', 'disallowed_lookups'):
            if option_name not in options:
                options[option_name] = getattr(self, f'models_{option_name}')
        model_config = ModelConfig(orm_model=orm_model, schema=self, **options)
        self.models_configs.append(model_config)
        orm_model_manager_class = model_config.orm_model_manager.__class__
        if orm_model_manager_class not in self.orm_model_manager_classes:
            self.orm_model_manager_classes.append(orm_model_manager_class)

    def execute(self, query, variables=None, operation_name=None,
            authenticated_user=None,
            serializable_output=False):
        """
            Execute a GraphQL query within the schema.
        """
        # debug logging
        self._requests_logger.debug(json.dumps({
            'authenticated_user': authenticated_user.username if authenticated_user is not None else None,
            'query': query,
            'operation_name': operation_name,
            'variables': variables,
        }))
        # info logging
        simpler_query = re.sub(r'#[^\n]+', '', query)
        simpler_query = re.sub(r'\s+', ' ', simpler_query).strip()
        self._requests_logger.info(json.dumps({
            'authenticated_user': authenticated_user.username if authenticated_user is not None else None,
            'query': simpler_query[:64] + '...' if len(simpler_query) > 64 else '',
        }))
        # actual query execution
        result = graphql_sync(
            schema = self._get_graphql_schema(),
            source = query,
            variable_values = variables or {},
            operation_name = operation_name,
            context_value = ContextValue(
                authenticated_user = authenticated_user,
            ),
        )
        # must the output be serializable?
        if serializable_output:
            formatted_result = result.formatted
            if 'errors' in formatted_result and not formatted_result['errors']:
                del formatted_result['errors']
            return formatted_result
        return result

    def check(self, graphql_schema=None):
        """
            Check the schema's validity; an exception is raised if something is wrong.
        """
        if graphql_schema is None:
            graphql_schema = self._get_graphql_schema()
        for error in validate_schema(graphql_schema):
            raise error

    def get_model_config(self, orm_model=None, name=None):
        """
            Retrieve an instance of `ModelConfig` if the passed `orm_model` has been
            exposed, `None` otherwise.
        """
        if not bool(orm_model) ^ bool(name):
            raise ValueError('You have to specify exactly one of the following parameters: '
                '`orm_model`, `name`')
        for model_config in self.models_configs:
            if model_config.orm_model_manager.orm_model == orm_model or model_config.name == name:
                return model_config
        return None

    def as_django_view(self, with_graphiql=True, compute_user=True):
        """
            Expose schema as a Django view.

            Example:

            ```python
            from django.urls import path
            import easy_graphql_server

            schema = easy_graphql_server.Schema()

            urlpatterns = [
                path('graphql', schema.as_django_view()),
            ]
            ```
        """
        # pylint: disable=import-outside-toplevel
        from .webserver.django_schema_view import DjangoSchemaView
        return DjangoSchemaView(
            schema = self,
            with_graphiql = with_graphiql,
            compute_user = compute_user
        ).view

    def as_flask_view(self):
        """
            Expose schema as a Flask view.

            Example:

            ```python
            from flask import Flask
            import easy_graphql_server

            schema = easy_graphql_server.Schema()

            app = Flask()

            app.add_url_rule(
                rule = '/graphql',
                endpoint = 'graphql',
                view_func = schema.as_flask_view())
            ```
        """
        # pylint: disable=import-outside-toplevel
        from .webserver.flask_schema_view import FlaskSchemaView
        return FlaskSchemaView(schema=self).view

    # private attributes & methods

    def _expose_method(self, type_, name, method, input_format=None, output_format=None,
            pass_graphql_selection=False, pass_graphql_path=False,
            pass_authenticated_user=False, require_authenticated_user=False, deprecation_message=None):
        self.methods[type_][name] = GraphQLField(
            # output format
            type_ = to_graphql_type(
                type_ = output_format,
                prefix = name,
                for_input = False,
                schema = self,
            ) if output_format else None,
            # input format
            args = to_graphql_argument(
                type_ = input_format,
                prefix = name,
                schema = self,
            ) if input_format else None,
            # resolve method
            resolve = self._make_callback(
                type_ = type_,
                method = method,
                pass_graphql_selection = (
                    'graphql_selection'
                    if pass_graphql_selection is True else
                    pass_graphql_selection
                ),
                pass_graphql_path = (
                    'graphql_path'
                    if pass_graphql_path is True else
                    pass_graphql_path
                ),
                pass_authenticated_user = (
                    'authenticated_user'
                    if pass_authenticated_user is True else
                    pass_authenticated_user
                ),
                require_authenticated_user = require_authenticated_user,
            ),
            # deprecation
            deprecation_reason = deprecation_message,
        )
        # schema is not up to date anymore
        self.dirty = True

    def _expose_from_subclass(self, subclass):
        subclass_attributes = introspection.get_public_class_attributes(subclass)
        # different parent classes, different results
        if issubclass(subclass, exposition.ExposedModel):
            message = ' when subclassing `ExposedModel`'
            exposition_method = self.expose_model
            validation_method = ModelConfig
            excluded_arguments = ('self', 'schema')
        elif issubclass(subclass, exposition.ExposedQuery):
            message = ' when subclassing `ExposedQuery`'
            exposition_method = self.expose_query
            validation_method = self._expose_method
            excluded_arguments = ('self', 'type_')
        elif issubclass(subclass, exposition.ExposedMutation):
            message = ' when subclassing `ExposedMutation`'
            exposition_method = self.expose_mutation
            validation_method = self._expose_method
            excluded_arguments = ('self', 'type_')
        else:
            raise ValueError('Parameter to `expose()` method should be a subclass of either '
                '`exposition.ExposedQuery`, `exposition.ExposedMutation` or '
                f'`exposition.ExposedModel`, but `{subclass}` was found instead')
        # attributes validation
        introspection.validate_class_attributes_against_method_arguments(
            cls = subclass,
            method = validation_method,
            excluded_arguments = excluded_arguments,
            message = message)
        # actual exposition
        exposition_method(**subclass_attributes)

    def _collect_from_classes(self):
        for subclass in introspection.get_subclasses(self.Exposed):
            if subclass not in (self.ExposedModel, self.ExposedQuery, self.ExposedMutation):
                self.expose(subclass)
        for subclass in self.subclasses:
            self._expose_from_subclass(subclass)


    # build schema when modified, returned cached version otherwise

    def _make_graphql_schema(self):
        self._collect_from_classes()
        # collect methods from exposed models
        for model_config in self.models_configs:
            model_config.expose_methods()
        # build and return schema
        return GraphQLSchema(
            query = GraphQLObjectType('Query',
                lambda: self.methods['query']) if self.methods['query'] else None,
            mutation = GraphQLObjectType('Mutation',
                lambda: self.methods['mutation']) if self.methods['query'] else None,
        )

    def _get_graphql_schema(self):
        if self.dirty:
            graphql_schema = self._make_graphql_schema()
            self.check(graphql_schema)
            self.graphql_schema = graphql_schema
            self.dirty = False
        return self.graphql_schema

    # build wrapper around passed methods to build a callback

    def _make_callback(self, type_, method,
            pass_graphql_selection, pass_graphql_path,
            pass_authenticated_user, require_authenticated_user):
        def callback(source, info, **kwargs): # pylint: disable=unused-argument
            try:
                # ensure authenticated user when mandatory
                if require_authenticated_user or pass_authenticated_user:
                    authenticated_user = info.context.authenticated_user
                if require_authenticated_user and not authenticated_user:
                    raise exceptions.UnauthenticatedError()
                # pass parameters
                if pass_authenticated_user:
                    kwargs[pass_authenticated_user] = authenticated_user
                if pass_graphql_selection:
                    kwargs[pass_graphql_selection] = self._get_graphql_selection(
                        info.field_nodes[0].selection_set, info.fragments)
                if pass_graphql_path:
                    kwargs[pass_graphql_path] = [type_, info.path.key]
                # executed method
                return method(**kwargs)
            except exceptions.BaseError as error:
                self._processing_logger.warning(error.args[0])
                raise
            except Exception as error:
                formatted_error = exceptions.InternalError(error)
                self._processing_logger.warning(formatted_error.args[0])
                if self.debug:
                    raise formatted_error
                raise exceptions.InternalError()
        return callback

    # using GraphQL core AST tree to return the GraphQL selection



    @classmethod
    def _get_graphql_selection(cls, selection_set, fragments, visited_fragments=None):
        """
            Return the GraphQL selection as a mapping.

            For example, with this query...

            ```gql
            query {
                houses {
                    location
                    tenants {
                        first_name
                        last_name
                    }
                }
            }
            ```

            ...the resulting mapping would be:

            ```python
            {
                'location': None,
                'tenants': {
                    'first_name': None,
                    'last_name': None,
                }
            }
            ```
        """
        # variables initialization
        if visited_fragments is None:
            visited_fragments = set()
        result = {}
        # browse selection
        for selection in selection_set.selections:
            if isinstance(selection, FieldNode):
                name = selection.alias.value if selection.alias else selection.name.value
                if name == '__typename':
                    continue
                result[name] = (
                    cls._get_graphql_selection(selection.selection_set, fragments, visited_fragments)
                    if getattr(selection, 'selection_set', None) else None)
            elif isinstance(selection, InlineFragmentNode):
                raise NotImplementedError()
            elif isinstance(selection, FragmentSpreadNode):
                fragment = fragments.get(selection.name.value)
                if not fragment or not hasattr(fragment, 'selection_set'):
                    continue
                result.update(
                    cls._get_graphql_selection(
                        fragment.selection_set, fragments, visited_fragments | {selection.name.value})
                )
        # the end!
        return result



"""
    Helper functions to convert (almost) anything into a GraphQL type, object type or argument.
"""


import inspect
import re
import enum
import datetime
import decimal

from graphql.type.definition import GraphQLType, GraphQLInputField # pylint: disable=
from graphql import \
    GraphQLEnumType, GraphQLEnumValue, \
    GraphQLField, GraphQLArgument, \
    GraphQLObjectType, GraphQLInputObjectType

from . import types, graphql_types, introspection


PYTHON_GRAPHQL_TYPES_MAPPING = {
    # boolean
    bool: graphql_types.Boolean,
    # integers
    int: graphql_types.Int,
    # other numbers
    float: graphql_types.Float,
    decimal.Decimal: graphql_types.Decimal,
    # texts
    str: graphql_types.String,
    # date
    datetime.date: graphql_types.Date,
    datetime.datetime: graphql_types.DateTime,
    datetime.time: graphql_types.Time,
}


_enums_cache = {}


def to_graphql_enum_key(name, capitalize=True):
    """
        Convert an enum name to its GraphQL-compatible counterpart

        (only uppercase letters, underscores and digits, cannot start with a digit)
    """
    key = re.sub(r'(?:^\d|[^a-zA-Z0-9])', '_', str(name))
    if capitalize:
        return key.upper()
    return key

def to_graphql_enum_from_choices(prefix, choices, description=None, capitalize=True, schema=None):
    # pylint: disable=unused-argument
    """
        Create a `GraphQLEnumType` from a list of choices.

        Choices must presented as a list of key-value pairs.
    """
    if prefix in _enums_cache:
        graphql_type = _enums_cache[prefix]
    else:
        graphql_type = _enums_cache[prefix] = GraphQLEnumType(f'{prefix}__enum_type', {
            to_graphql_enum_key(key, capitalize): GraphQLEnumValue(key, value)
            for key, value in choices
        }, description=description)
    return graphql_type

def to_graphql_enum_from_enum(prefix, enum):
    """
        Create a `GraphQLEnumType` from a subclass of native Python Enum.
    """
    pass

def to_graphql_type(type_, prefix, for_input=False, schema=None):
    """
        Returns a GraphQL type given a `GraphQLType`, a Python native type,
        a mapping, a single-item list or a subclass of `enum.Enum`.
    """
    # if this is already a graphql type, return it as is
    if introspection.is_instance_or_subclass(type_, GraphQLType):
        return type_
    # mapping
    if isinstance(type_, dict):
        return to_graphql_objecttype(type_, prefix, for_input=for_input, schema=schema)
    if inspect.isclass(type_):
        # convert native Python types
        if type_ in PYTHON_GRAPHQL_TYPES_MAPPING:
            return PYTHON_GRAPHQL_TYPES_MAPPING[type_]
        # enum
        if issubclass(type_, enum.Enum):
            return to_graphql_enum_from_choices(
                prefix = prefix,
                choices = [(key, value.value) for key, value in type_.__members__.items()],
                schema = schema)
    # list
    if isinstance(type_, list) and len(type_) == 1:
        return graphql_types.List(
            to_graphql_type(type_[0], prefix, for_input=for_input, schema=schema)
        )
    # mandatory
    if isinstance(type_, types.Required):
        return graphql_types.NonNull(
            to_graphql_type(type_.type_, prefix, for_input=for_input, schema=schema)
        )
    # existing model interface
    if isinstance(type_, types.ModelInterface):
        model_config = schema.get_model_config(name=type_.model_name)
        if model_config is None:
            raise ValueError(f'No model in schema with name `{type_.model_name}` '
                '(consider changing the order of expositions declaration)')
        mapping = model_config.get_type_mapping(
            operation = type_.operation,
            with_custom_fields = True)
        if not type_.exclude and not type_.additional:
            return to_graphql_type(
                type_ = mapping,
                prefix = model_config.name,
                for_input = for_input,
                schema = schema)
        for field_name in type_.exclude:
            mapping.pop(field_name, None)
        mapping.update(type_.additional)
        return to_graphql_type(
            type_ = mapping,
            prefix = prefix,
            for_input = for_input,
            schema = schema)
    # existing model field
    if isinstance(type_, types.ModelField):
        model_config = schema.get_model_config(name=type_.model_name)
        if model_config is None:
            raise ValueError(f'No model in schema with name `{type_.model_name}` '
                '(consider changing the order of expositions declaration)')
        mapping = model_config.get_type_mapping(with_custom_fields = True)
        try:
            for field_name in type_.field_path:
                mapping = mapping[field_name]
        except KeyError as error:
            raise ValueError(
                f'Invalid field path `{".".join(type_.field_path)}` '
                f'for model named `{type_.model_name}`') from error
        return to_graphql_type(
            type_ = mapping,
            prefix = prefix,
            for_input = for_input,
            schema = schema)
    # oops.
    if isinstance(type_, types.Model):
        raise ValueError('Could not convert `easy_graphql_server.Model` instance '
            'to graphql type, use instead `Model(...).output_format`, '
            '`Model(...).create_input_format` or `Model(...).update_input_format`')
    raise ValueError(f'Could not convert {type_} to graphql type')


_objecttype_cache = {}

def to_graphql_objecttype(type_, prefix, for_input=False, schema=None):
    """
        Returns a `GraphQLInputObjectType` or a `GraphQLObjectType` from a given type.

        Possible input `type_` are `GraphQLObjectType`, or a `dict` mapping of strings
        to types that are supported by `to_graphql_type()`.

        Function returns a `GraphQLObjectType` that uses `GraphQLField` when `for_input == False`,
        a `GraphQLInputObjectType` with `GraphQLInputField` when `for_input == True`.
    """
    # already a field
    if isinstance(type_, GraphQLObjectType):
        return type_
    # in cache
    cache_key = (prefix, for_input)
    if cache_key in _objecttype_cache:
        return _objecttype_cache[cache_key]
    # mapping
    if isinstance(type_, dict):
        if for_input:
            object_type_class = GraphQLInputObjectType
            object_name = f'{prefix}__input_type'
            field_class = GraphQLInputField
        else:
            object_type_class = GraphQLObjectType
            object_name = f'{prefix}__output_type'
            field_class = GraphQLField
        object_type = object_type_class(object_name, lambda : {
            key: field_class(to_graphql_type(
                type_ = value,
                prefix = f'{prefix}__{key}',
                for_input = for_input,
                schema = schema))
            for key, value in type_.items()
        })
        _objecttype_cache[cache_key] = object_type
        return object_type
    # otherwise
    return GraphQLInputField(type_) if for_input else GraphQLField(type_)

def to_graphql_argument(type_, prefix, schema=None):
    """
        Returns a `GraphQLArgument`, taking as argument either a `GraphQLArgument`
        or `dict` mapping of `str` to types that are supported by `to_graphql_type()`.
    """
    # already a field
    if isinstance(type_, GraphQLArgument):
        return type_
    # mapping
    if isinstance(type_, dict):
        return {
            key: (
                value
                if isinstance(value, GraphQLArgument) else
                GraphQLArgument(to_graphql_type(
                    type_ = value,
                    prefix = f'{prefix}__{key}',
                    for_input = True,
                    schema = schema))
            )
            for key, value in type_.items()
        }
    # otherwise
    raise ValueError(f'Could not convert `{type_}` of type `{type(type_)}` to GraphQL argument, '
        'expecting either a instance of `GraphQLArgument` or `dict`')



"""
    This module defines the `ModelConfig` class.
"""

from collections import defaultdict

from .operations import Operation
from .conversion import to_graphql_objecttype, to_graphql_argument
from .types import Required
from .orm import ORM
from .model_config_custom_field import ModelConfigCustomField
from . import exceptions, introspection, graphql_types
from .exposition import CustomField


class ModelConfig:

    """
        Configuration of an exposed model.

        Internally used within `Schema` class as a `Schema.models_configs` collection
        to store configuration for exposed models.
    """

    def __init__(self, schema, orm_model,
            name=None, plural_name=None, types_name=None,
            can_expose=True, cannot_expose=False,
            can_create=True, can_read=True, can_update=True, can_write=True, can_delete=True,
            cannot_create=False, cannot_update=False, cannot_read=False, cannot_write=False,
            cannot_delete=False,
            only_when_child_of=None, require_authenticated_user=False, restrict_queried_fields=False,
            has_permission=None, filter_for_user=None,
            on_before_operation=None, on_after_operation=None,
            allowed_lookups=None, disallowed_lookups=None,
            custom_fields=None, max_depth=None, limit=-1):
        # pylint: disable=unused-argument # for callbacks

        # store raw options
        self.schema = schema
        if require_authenticated_user is False:
            self.require_authenticated_user = ()
        elif require_authenticated_user is True:
            self.require_authenticated_user = (Operation.CREATE, Operation.READ, Operation.UPDATE, Operation.DELETE)
        else:
            self.require_authenticated_user = require_authenticated_user
        self.only_when_child_of = only_when_child_of
        self.max_depth = max_depth
        self.limit = limit
        # callbacks
        callbacks_names = ('has_permission', 'filter_for_user', 'on_before_operation', 'on_after_operation')
        self.callbacks = defaultdict(list)
        for callback_name in callbacks_names:
            local_callback = locals().get(callback_name)
            if local_callback:
                self.callbacks[callback_name].append(local_callback)
            model_callback_name = f'egs_{callback_name}'
            model_callback = orm_model.__dict__.get(model_callback_name)
            if model_callback:
                if isinstance(model_callback, (staticmethod, classmethod)):
                    callback = getattr(orm_model, model_callback_name)
                    self.callbacks[callback_name].append(callback)
                else:
                    def make_callback(model_callback_name):
                        return (lambda instance, *args, **kwargs:
                            getattr(instance, model_callback_name)(*args, **kwargs))
                    self.callbacks[callback_name].append(
                        make_callback(model_callback_name))
        # custom fields
        self.custom_fields = []
        if custom_fields:
            for custom_field in custom_fields:
                if isinstance(custom_field, dict):
                    self._add_custom_field(**custom_field)
                elif issubclass(custom_field, CustomField):
                    introspection.validate_class_attributes_against_method_arguments(
                        cls = custom_field,
                        method = ModelConfigCustomField,
                        excluded_arguments = ('self', ))
                    attributes = introspection.get_public_class_attributes(custom_field)
                    self._add_custom_field(**attributes)
        # what methods are exposed
        self.available_operations = {
            Operation.READ: (can_read is not False and cannot_read is not True),
            Operation.CREATE: (can_write is not False and
                can_create is not False and cannot_create is not True),
            Operation.UPDATE: (can_write is not False and
                can_update is not False and cannot_update is not True),
            Operation.DELETE: (can_write is not False and
                can_delete is not False and cannot_delete is not True),
        }
        # what fields are exposed in which method
        self.concatenated_fields = {
            Operation.CREATE: self._concatenate_fields_options(
                (can_expose, can_write, can_create), True),
            Operation.READ: self._concatenate_fields_options(
                (can_expose, can_read,), True),
            Operation.UPDATE: self._concatenate_fields_options(
                (can_expose, can_write, can_update), True),
        }
        self.concatenated_exclude = {
            Operation.CREATE: self._concatenate_fields_options(
                (cannot_expose, cannot_write, cannot_create), False),
            Operation.READ: self._concatenate_fields_options(
                (cannot_expose, cannot_read,), False),
            Operation.UPDATE: self._concatenate_fields_options(
                (cannot_expose, cannot_write, cannot_update), False),
        }
        # instanciate ORM model manager
        self.orm_model_manager = ORM.get_manager(
            orm_model = orm_model,
            model_config = self,
            restrict_queried_fields = restrict_queried_fields,
        )
        # name
        self.name = name or schema.case_manager.convert(self.orm_model_manager.get_table_name())
        self.types_name = types_name or self.name
        self.plural_name = plural_name or f'{self.name}s'
        # lookups
        self.allowed_lookups = allowed_lookups
        self.disallowed_lookups = disallowed_lookups or ()

    def expose_methods(self):
        """
            Called from `Schema` just before generating actual GraphQL schema.
            Exposes methods to read, write, create, and update model instances.

            Only allowed methods will be exposed.
        """
        # do not expose if explicitely forbidden
        if self.only_when_child_of:
            return
        # available filters for querying
        filters = self.orm_model_manager.get_filters()
        # this is the common output format for all methods
        output_type = to_graphql_objecttype(
            type_ = self.get_type_mapping(Operation.READ, require_non_nullable=True),
            prefix = self.types_name,
            schema = self.schema,
        )
        # expose create method
        if self.available_operations[Operation.CREATE]:
            # create one instance
            self.schema.expose_mutation(
                name = f'create_{self.name}',
                input_format = to_graphql_argument(
                    type_ = self.get_type_mapping(Operation.CREATE),
                    prefix = f'create_{self.name}',
                    schema = self.schema,
                ),
                output_format = output_type,
                method = self.orm_model_manager.decorate(
                    self.orm_model_manager.create_one),
                pass_graphql_path = True,
                pass_graphql_selection = True,
                pass_authenticated_user = True,
                require_authenticated_user = Operation.CREATE in self.require_authenticated_user,
            )
        # expose read methods
        if self.available_operations[Operation.READ]:
            # fetch one instance
            self.schema.expose_query(
                name = self.name,
                input_format = self.orm_model_manager.fields_info.unique,
                output_format = output_type,
                method = self.orm_model_manager.decorate(
                    self.orm_model_manager.read_one),
                pass_graphql_path = True,
                pass_graphql_selection = True,
                pass_authenticated_user = True,
                require_authenticated_user = Operation.READ in self.require_authenticated_user,
            )
            # fetch many instances
            self.schema.expose_query(
                name = self.plural_name,
                input_format = filters,
                output_format = [output_type],
                method = self.orm_model_manager.decorate(
                    self.orm_model_manager.read_many),
                pass_graphql_path = True,
                pass_graphql_selection = True,
                pass_authenticated_user = True,
                require_authenticated_user = Operation.READ in self.require_authenticated_user,
            )
        # expose UPDATE method
        if self.available_operations[Operation.UPDATE]:
            # update one instance
            self.schema.expose_mutation(
                name = f'update_{self.name}',
                input_format = to_graphql_argument(
                    type_ = dict(
                        {'_': self.get_type_mapping(Operation.UPDATE)},
                        ** self.orm_model_manager.fields_info.unique),
                    prefix = f'update_{self.name}',
                    schema = self.schema,
                ),
                output_format = output_type,
                method = self.orm_model_manager.decorate(
                    self.orm_model_manager.update_one),
                pass_graphql_path = True,
                pass_graphql_selection = True,
                pass_authenticated_user = True,
                require_authenticated_user = Operation.UPDATE in self.require_authenticated_user,
            )
        # expose delete method
        if self.available_operations[Operation.DELETE]:
            # delete one instance
            self.schema.expose_mutation(
                name = f'delete_{self.name}',
                input_format = self.orm_model_manager.fields_info.unique,
                output_format = output_type,
                method = self.orm_model_manager.decorate(
                    self.orm_model_manager.delete_one),
                pass_graphql_path = True,
                pass_graphql_selection = True,
                pass_authenticated_user = True,
                require_authenticated_user = Operation.DELETE in self.require_authenticated_user,
            )

    # concatenate

    @staticmethod
    def _concatenate_fields_options(options, continue_if):
        concatenated = set()
        for option in options:
            # if option is a boolean
            if option is continue_if:
                continue
            if option is (not continue_if):
                return None
            # if option is an iterator
            if concatenated is None:
                concatenated = set(option)
            else:
                concatenated |= set(option)
        return tuple(concatenated) if concatenated else None

    # types computation

    def get_type_mapping(self, operation=None, exclude=None, depth=0, linked_field=None,
            with_custom_fields=True, max_depth=None, require_non_nullable=False):
        """
            Return a `dict` from `str` to GraphQL types, corresponding to the model.

            Using recursion, it allows nesting, in a way that we never go through a given foreign
            key twice.
        """
        fields_info = self.orm_model_manager.fields_info
        exclude = exclude or set()
        mapping = {}
        # check depth
        if self.max_depth is not None and (max_depth is None or self.max_depth < max_depth):
            max_depth = self.max_depth
        # no mapping for deletion
        if operation == Operation.DELETE:
            return {}
        # value fields
        for field_name, graphql_type in fields_info.value.items():
            # primary fields mustn't be exposed for creation & update, unless nested
            if operation in (Operation.CREATE, Operation.UPDATE):
                if depth == 0 and field_name == fields_info.primary:
                    continue
            # ensure the field is exposed for this operation
            if operation is not None and not self.can_perform(operation, field_name):
                continue
            # map
            mapping[field_name] = graphql_type
        # foreign & related fields...
        if max_depth is None or max_depth > 0:
            for field_name, field in fields_info.linked.items():
                # ensure the field is exposed for this operation
                if operation is not None and not self.can_perform(operation, field_name):
                    continue
                # retrieve other model config
                other_model_config = self.schema.get_model_config(orm_model=field.orm_model)
                if other_model_config is None:
                    continue
                # only_when_child_of is important
                if other_model_config.only_when_child_of and not issubclass(
                        self.orm_model_manager.orm_model, other_model_config.only_when_child_of):
                    continue
                # these fields are at stake, and will be later excluded
                _exclude = {(self, field_name), (other_model_config, field.field_name)}
                if exclude & _exclude:
                    continue
                # ...with recursion
                mapping[field_name] = other_model_config.get_type_mapping(
                    operation = operation,
                    exclude = exclude | _exclude,
                    depth = depth + 1,
                    max_depth = (max_depth - 1) if max_depth is not None else None,
                    linked_field = field,
                    with_custom_fields = with_custom_fields,
                    require_non_nullable = require_non_nullable)
                # related are presented as collections
                if field_name in fields_info.related:
                    if require_non_nullable:
                        mapping[field_name] = Required([Required(mapping[field_name])])
                    else:
                        mapping[field_name] = [mapping[field_name]]
        # apply non null when necessary
        if operation == Operation.CREATE:
            for field_name, graphql_type in list(mapping.items()):
                if field_name in fields_info.mandatory:
                    if linked_field is not None and field_name == linked_field.value_field_name:
                        continue
                    mapping[field_name] = self._make_required(graphql_type)
        if require_non_nullable:
            for field_name, graphql_type in list(mapping.items()):
                if field_name not in fields_info.nullable:
                    mapping[field_name] = self._make_required(graphql_type)
        # custom fields
        if with_custom_fields:
            for custom_field in self.custom_fields:
                if operation is None or custom_field.can_perfom(operation):
                    mapping[custom_field.name] = custom_field.format
        # result
        return mapping

    # required

    @classmethod
    def _make_required(cls, graphql_type):
        if isinstance(graphql_type, (graphql_types.NonNull, Required)):
            return graphql_type
        if isinstance(graphql_type, graphql_types.List):
            return graphql_types.NonNull(graphql_types.List(graphql_types.NonNull(graphql_type.of_type)))
        if isinstance(graphql_type, list):
            return Required([Required(graphql_type[0])])
        return Required(graphql_type)

    # callbacks

    def _call(self, callback_name, *args, **kwargs):
        for method in self.callbacks[callback_name]:
            yield method(*args, **kwargs)

    # authorizations

    def can_perform(self, operation, field_name):
        """
            Check if a given `Operation` can be performed on a given field, given its `str` name.

            Of course, this does not make sense for `Operation.DELETE`, which raises an exception.
        """
        # deletion doesn't apply to one field
        if operation == Operation.DELETE:
            raise ValueError(
                'Model.can_perform can only be called with `operation != Operation.DELETE`')
        # cannot if not in explicitely exposed fields
        if self.concatenated_fields[operation]:
            if field_name not in self.concatenated_fields[operation]:
                return False
        # cannot if in explicitely forbidden fields
        if self.concatenated_exclude[operation]:
            if field_name in self.concatenated_exclude[operation]:
                return False
        # otherwise, it works
        return True

    def can_expose_from_parent(self, orm_model):
        """
            Check if the present ORM model can be exposed when nested under another model.

            This is useful when a model cannot be exposed as a standalone, but only
            nested under some other model.
        """
        if not self.only_when_child_of:
            return True
        return issubclass(orm_model, self.only_when_child_of)

    def filter_for_user(self, queryset, authenticated_user):
        """
            Returns a Django `Queryset`, which is a filtered version of the input `queryset`
            only showing what is available to the `authenticated_user`.
        """
        for filter_for_user_method in self.callbacks['filter_for_user']:
            queryset = filter_for_user_method(queryset, authenticated_user)
        return queryset

    def has_permission(self, instance, authenticated_user, operation, data=None):
        """
            Return a boolean indicating whether or not the requested operation can
            be performed on the instance.

            `data` is provided when the operation is either `Operation.CREATE` or
            `Operation.UPDATE`. It is a dictionary containing the new provided data
            for the instance.
        """
        if data is None:
            data = {}
        return all(self._call('has_permission', instance, authenticated_user, operation, data))

    def ensure_permission(self, instance, authenticated_user, operation,
            data=None, graphql_path=None):
        """
            Raise an `exceptions.ForbiddenError` when `has_permission()` returns
            `False` with the same parameters, aka when the requested operation cannot
            be performed on the instance.

            `data` is provided when the operation is either `Operation.CREATE` or
            `Operation.UPDATE`. It is a dictionary containing the new provided data
            for the instance.

            The `graphql_path` parameter indicates where the permission was denied.
        """
        if data is None:
            data = {}
        if not self.has_permission(instance, authenticated_user, operation, data):
            raise exceptions.ForbiddenError(
                operation = operation,
                authenticated_user = authenticated_user,
                path = graphql_path,
            )

    # lookups

    def is_lookup_allowed(self, name):
        if self.allowed_lookups is None or name in self.allowed_lookups:
            return name not in self.disallowed_lookups
        return False

    # triggers

    def on_before_operation(self, instance, authenticated_user, operation, data=None, depth=0):
        """
            Execute pre-trigger(s) for given instance with given parameters.
        """
        if not isinstance(instance, self.orm_model_manager.orm_model):
            raise Exception(f'Unexpected instance {instance}, should be of type {self.orm_model_manager.orm_model}')
        return list(self._call('on_before_operation', instance, authenticated_user, operation, data, depth))

    def on_after_operation(self, instance, authenticated_user, operation, data=None, depth=0):
        """
            Execute post-trigger(s) for given instance with given parameters.
        """
        if not isinstance(instance, self.orm_model_manager.orm_model):
            raise Exception(f'Unexpected instance {instance}, should be of type {self.orm_model_manager.orm_model}')
        return list(self._call('on_after_operation', instance, authenticated_user, operation, data, depth))

    # custom field

    def _add_custom_field(self, *args, **kwargs):
        custom_field = ModelConfigCustomField(*args, **kwargs)
        self.custom_fields.append(custom_field)



"""
    Definition of `SqlAlchemyModelManager` class.
"""

from ._manager import ModelManager


# pylint:disable=W0223 # Methods are abstract in class 'ModelManager' but not overridden


class SqlAlchemyModelManager(ModelManager):
    """
        ModelManager class for SqlAlchemy ORM.
    """

    # remove Pylint disabling above class declaration when overridding methods



"""
    This module defines the `LOOKUPS` constant, useful for defining filters on fields.

    This is currently used for Django lookups, but could be very interesting for other ORMs too.
"""


from .. import graphql_types


LOOKUPS = {

    graphql_types.Boolean: {},

    graphql_types.Int: {},

    graphql_types.Float: {},

    graphql_types.Decimal: {},

    graphql_types.String: {
        'iexact': graphql_types.String,
        'contains': graphql_types.String,
        'icontains': graphql_types.String,
        'startswith': graphql_types.String,
        'istartswith': graphql_types.String,
        'endswith': graphql_types.String,
        'iendswith': graphql_types.String,
        'regex': graphql_types.String,
        'iregex': graphql_types.String,
    },

    graphql_types.DateTime: {
        'date': graphql_types.Date,
        'time': graphql_types.Time,
    },

    graphql_types.Date: {
        'year': graphql_types.Int,
        'iso_year': graphql_types.Int,
        'quarter': graphql_types.Int,
        'month': graphql_types.Int,
        'week': graphql_types.Int,
        'iso_week_day': graphql_types.Int,
        'week_day': graphql_types.Int,
        'day': graphql_types.Int,
    },

    graphql_types.Time: {
        'hour': graphql_types.Int,
        'minute': graphql_types.Int,
        'second': graphql_types.Int,
    },

}


LOOKUPS[graphql_types.DateTime].update(LOOKUPS[graphql_types.Date])
LOOKUPS[graphql_types.DateTime].update(LOOKUPS[graphql_types.Time])

for graphql_type in list(LOOKUPS.keys()):
    if graphql_type == graphql_types.Boolean:
        continue
    LOOKUPS[graphql_type].update({
        'lt': graphql_type,
        'lte': graphql_type,
        'gt': graphql_type,
        'gte': graphql_type,
        'in': graphql_types.List(graphql_type),
        'isnull': graphql_types.Boolean,
    })



"""
    Definition of `DjangoSchemaView` class
"""

from django.http import HttpResponse, JsonResponse
import django.contrib.auth
from ._schema_view import SchemaView


class DjangoSchemaView(SchemaView):

    """
        Django schema view. Used when calling `Schema.as_django_view()`.
    """

    def __init__(self, *args, **kwargs):
        self.compute_user = kwargs.pop('compute_user', True)
        super().__init__(*args, **kwargs)

    def view(self, request):
        """
            Django view to compute GraphQL request
        """
        # compute user when requested and not present
        authenticated_user = request.user
        if self.compute_user:
            if not authenticated_user or not authenticated_user.is_authenticated or authenticated_user.is_anonymous:
                try:
                    authenticated_user = django.contrib.auth.authenticate(request)
                except Exception: # pylint: disable=broad-except
                    pass
        # compute result
        result = self.compute_response(
            method = request.method,
            headers = request.headers,
            body = request.body,
            query = dict(request.GET.items()),
            authenticated_user = authenticated_user,
        )
        # return response
        if isinstance(result, str):
            return HttpResponse(result)
        return JsonResponse(result[0], status=result[1])



"""
    Definition of easily serializable exceptions, meant to be thrown from the GraphQL API.
"""

import re

from graphql import GraphQLError
from graphql.language.ast import Node

from . import custom_json


class BaseError(Exception):
    """
        Base exception for `easy_graphql_server`
    """
    def __init__(self, type_, payload):
        message = custom_json.dumps({
            'type': type_,
            'payload': payload,
        })
        Exception.__init__(self, message)


_graphqlerror__init__ = GraphQLError.__init__
def graphqlerror__init__(self, message, nodes=None, source=None, positions=None, path=None,
        original_error=None, extensions=None):
    """
        Patch for the constructor of GraphQLError class, defined in graphql-core.

        Gives more details, in the same format as BaseError output.
    """
    if original_error is None or not isinstance(original_error, BaseError):
        type_ = None
        # parse original error message
        possible_errors = (
            ('MISSING_ARGUMENT',
                r"^Field '(?P<parent_type_name>.*?)\.(?P<expected_field_name>.*?)' of required type "
                r"'(?P<expected_type>.*?)' was not provided\."),
            ('MISSING_ARGUMENT',
                r"^Field '(?P<parent_type_name>.*?)' argument '(?P<expected_argument_name>.*?)' of type "
                r"'(?P<expected_type>.*?)' is required, but it was not provided\."),
            ('UNEXPECTED_ARGUMENT',
                r"^Field '(?P<unexpected_argument_name>.*?)' is not defined by type '(?P<parent_type_name>.*?)'\."),
            ('UNEXPECTED_ARGUMENT',
                r"^Unknown argument '(?P<unexpected_argument_name>.*?)' on field '(?:Mutation|Query)\.(?P<method_name>.*?)'\."),
            ('METHOD_NOT_FOUND',
                r"^Cannot query field '(?P<method_name>.*?)' on type '(?:Mutation|Query)'\.$"),
            ('METHOD_NOT_FOUND',
                r"^Cannot query field '(?P<method_name>.*?)' on type '(?:Mutation|Query)'\. Did you mean "
                r"(?P<suggestions>.*?)\?$"),
            ('UNEXPECTED_QUERIED_FIELD',
                r"^Cannot query field '(?P<unexpected_field_name>.*?)' on type '(?P<parent_type_name>.*?)'\.$"),
            ('UNEXPECTED_QUERIED_FIELD',
                r"^Cannot query field '(?P<unexpected_field_name>.*?)' on type '(?P<parent_type_name>.*?)'\. Did you "
                r"mean (?P<suggestions>.*?)\?$"),
            ('WRONGLY_TYPED_ARGUMENT',
                r"^Expected value of type '(?P<expected_field_type>.*?)', found (?P<provided_field_value>.*?)\."),
            ('NON_NULL_FIELD',
                r"^Cannot return null for non\-nullable field (?P<field_path>)\."),
            ('SYNTAX_ERROR',
                r"^Syntax error: (?P<message>.*?)$"),
            ('WRONG_ENUM_VALUE',
                r"^Value '(?P<provided_field_value>.*?)' does not exist in '(?P<enum_type>.*?)' enum\.$"),
            ('UNKNOWN_TYPE',
                r"^Unknown type '(?P<UNKNOWN_TYPE>.*?)'\."),
        )
        for possible_error, regex in possible_errors:
            match = re.search(regex, message)
            if match:
                type_ = possible_error
                payload = match.groupdict() or {}
                break
        if type_ is None:
            type_ = 'GRAPHQL'
            payload = {'message': message}
        # parse suggestions
        if type_ in ('UNEXPECTED_QUERIED_FIELD', 'METHOD_NOT_FOUND'):
            if 'suggestions' in payload:
                payload['suggestions'] = re.findall(r"'([^']+)'", payload['suggestions'])
            else:
                payload['suggestions'] = []
        # build path when possible
        if isinstance(nodes, Node):
            token = nodes.loc.start_token
            if token.kind.value == 'Name':
                path = [token.value]
            ready = False
            while token.prev:
                token = token.prev
                if token.kind.value in '({':
                    ready = True
                elif ready and token.kind.value == 'Name':
                    if path:
                        path.append(token.value)
                    else:
                        path = [token.value]
                    ready = False
            payload['path'] = path = path[::-1]
        # compute message for path
        message = custom_json.dumps({
            'type': type_,
            'payload': payload,
        })
    return _graphqlerror__init__(
        self, message, nodes, source, positions, path, original_error, extensions)
setattr(GraphQLError, '__init__', graphqlerror__init__)


class InternalError(BaseError):
    """
        Internal coding error
    """
    def __init__(self, exception=None):
        if isinstance(exception, Exception):
            # extract traceback
            tracebacks = []
            traceback = getattr(exception, '__traceback__')
            while traceback:
                tracebacks.append(traceback)
                traceback = traceback.tb_next
            # build payload
            exception_class = type(exception)
            payload = {
                'class': (
                    exception_class.__qualname__
                    if exception_class.__module__ == 'builtins' else
                    f'{exception_class.__module__}.{exception_class.__qualname__}'
                ),
                'args': list(map(str, exception.args)),
                'traceback': [
                    f'{traceback.tb_frame.f_code.co_filename}:{traceback.tb_lineno}'
                    for traceback in tracebacks
                ]
            }
        else:
            payload = {}
        BaseError.__init__(self, 'INTERNAL', payload)

class UnauthenticatedError(BaseError):
    """
        Thrown when authentication is required for a query, but not provided.
    """
    def __init__(self):
        BaseError.__init__(self, 'UNAUTHENTICATED', {})

class NotFoundError(BaseError):
    """
        Thrown when an item was not found in database.
    """
    def __init__(self, filters):
        BaseError.__init__(self, 'NOT_FOUND', {
            'filters': filters,
        })

class ForbiddenError(BaseError):
    """
        Thrown when an authenticated user is not allowed to perform an operation
        on a specific item.
    """
    def __init__(self, operation, authenticated_user, path):
        BaseError.__init__(self, 'FORBIDDEN', {
            'operation': operation.name,
            'authenticated_user': str(authenticated_user),
            'path': path,
        })

class ValidationError(BaseError):
    """
        Thrown when attempting to save wrong data into database.
    """
    def __init__(self, issues):
        BaseError.__init__(self, 'VALIDATION', issues)

class DuplicateError(BaseError):
    """
        Thrown when an item is seen as duplicate in database.
    """
    def __init__(self, path):
        BaseError.__init__(self, 'DUPLICATE', {
            'path': path,
        })

class IntegrityError(BaseError):
    """
        Thrown when an item cannot be removed without compromising the database.
    """
    def __init__(self, path):
        BaseError.__init__(self, 'INTEGRITY', {
            'path': path,
        })



"""
    Definition of `ForeignField`, `RelatedField` and, most importantly, `FieldsInfo`.
"""


class LinkedField:
    # pylint: disable=too-few-public-methods
    """
        Base class for `ForeignField` and `RelatedField`.
    """

    def __init__(self, orm_model, value_field_name, field_name):
        self.orm_model = orm_model
        self.field_name = field_name
        self.value_field_name = value_field_name

    def __repr__(self):
        return (f'<{self.__class__.__name__} orm_model={self.orm_model.__name__} '
            f'field_name={self.field_name} value_field_name={self.value_field_name}>')


class ForeignField(LinkedField): # pylint: disable=too-few-public-methods
    """
        Description of a foreign key field.

        `orm_model` is the ORM model to which the foreign key is pointing.

        `field_name` is the field name on the ORM model that refers the foreign instance.

        `value_field_name` is the actual foreign key, bearing the value of the
        primary key on the other model.
    """

class RelatedField(LinkedField): # pylint: disable=too-few-public-methods
    """
        Description of a "related" field (aka., the opposite of a foreign key field).

        `orm_model` is the ORM model from which the foreign key is pointing.

        `field_name` is the field name on the other ORM model that refers the given instance.

        `value_field_name` is the actual foreign key on the other ORM model, bearing the value
        of the primary key on the given model.
    """


class FieldsInfo: # pylint: disable=too-few-public-methods
    """
        Info about fields for a given ORM model.
    """
    def __init__(self):
        # primary key as `str`
        self.primary = None
        # unique keys as `dict[str,GraphQLScalarType]`
        self.unique = {}
        # value fields as `dict[str,GraphQLType]`
        self.value = {}
        # foreign fields as `dict[str,ForeignField]`
        self.foreign = {}
        # related fields as `dict[str,RelatedField]`
        self.related = {}
        # linked fields as `dict[str,LinkedField]` (self.foreign | self.related)
        self.linked = None
        # mandatory fields upon creation as `set[str]`
        self.mandatory = set()
        # mandatory fields upon creation as `set[str]`
        self.nullable = set()
        # mandatory fields upon creation as `set[str]`
        self.custom = set()

    def compute_linked(self):
        """
            Merge `self.foreign` and `self.related` into `self.linked`
        """
        self.linked = dict(self.foreign, **self.related)



"""
    Definition of ORM managers.
"""

import enum
import inspect
import re


@enum.unique
class ORM(enum.Enum):
    """
        Enumeration of available ORM managers.

        Only Django is supported so far.
    """

    UNKNOWN = 0
    DJANGO = 1
    SQLALCHEMY = 2
    PEEWEE = 3

    @classmethod
    def identify(cls, orm_model):
        """
            Identify which ORM the passed model belongs to.
        """
        if inspect.isclass(orm_model):
            inherited_classes_names = tuple(map(
                lambda cls: re.sub(r'^builtins\.', '', f'{cls.__module__}.{cls.__name__}'),
                orm_model.__mro__))
            for inherited_class_name in inherited_classes_names:
                if inherited_class_name.startswith('django.'):
                    return cls.DJANGO
                if inherited_class_name.startswith('peewee.'):
                    return cls.PEEWEE
                if inherited_class_name.startswith('sqlalchemy.'):
                    return cls.SQLALCHEMY
        return cls.UNKNOWN

    @classmethod
    def get_manager_class(cls, orm_model):
        """
            Returns manager class corresponding to a given model.
        """
        # pylint: disable=import-outside-toplevel
        # managers are directly imported below `if` clauses, so we don't have to install
        # all ORMs for `easy_graphql_server` to function
        orm = cls.identify(orm_model)
        if orm == cls.UNKNOWN:
            raise ValueError(
                f'Unrecognized ORM for model, is neither Django, SQLAlchemy or Peewee: {orm_model}')
        if orm == cls.DJANGO:
            from .django_manager import DjangoModelManager
            return DjangoModelManager
        if orm == cls.SQLALCHEMY:
            from .sqlalchemy_manager import SqlAlchemyModelManager
            return SqlAlchemyModelManager
        if orm == cls.PEEWEE:
            from .peewee_manager import PeeweeModelManager
            return PeeweeModelManager
        raise ValueError('This should not happen.')

    @classmethod
    def get_manager(cls, orm_model, model_config, restrict_queried_fields):
        """
            Returns manager object corresponding to a given model.
        """
        manager_class = cls.get_manager_class(orm_model)
        return manager_class(orm_model, model_config, restrict_queried_fields)



"""
    Definition of `ModelManager` base class.
"""

from .. import graphql_types
from ..operations import Operation
from ._lookups import LOOKUPS

import graphql.type.definition


class ModelManager:
    """
        Base class for `DjangoModelManager`, `SqlAlchemyModelManager`, `PeeweeModelManager`

        Cannot be used as is, this is an abstract class that must be inherited.
    """

    def __init__(self, orm_model, model_config, restrict_queried_fields):
        self.orm_model = orm_model
        self.model_config = model_config
        self.restrict_queried_fields = restrict_queried_fields

    @property
    def fields_info(self):
        if not hasattr(self, '_fields_info'):
            self._fields_info = self.get_fields_info()
            self._fields_info.compute_linked()
            for custom_field in self.model_config.custom_fields:
                self._fields_info.custom.add(custom_field.name)
        return self._fields_info

    # metadata extraction

    def get_fields_info(self):
        """
            Retrieve fields info for the given ORM model.
        """
        raise NotImplementedError()

    def get_filters(self, mapping=None, prefix=''):
        """
            Retrieve available filters for the given ORM model.

            Needs corresponding `easy_graphql_server.ModelConfig` to be properly initialize, aka.,
            just before GraphQL schema is built.
        """
        filters = {}
        # base mapping
        if mapping is None:
            mapping = self.model_config.get_type_mapping(
                operation = Operation.READ,
                with_custom_fields = False)
        # browse all fields
        for field_name, graphql_type in mapping.items():
            prefixed_field_name = f'{prefix}{field_name}'
            # foreign & related field
            if isinstance(graphql_type, (dict, list)):
                submapping = graphql_type[0] if isinstance(graphql_type, list) else graphql_type
                filters.update(self.get_filters(submapping, f'{prefixed_field_name}__'))
            # value field
            elif graphql_type in LOOKUPS:
                # basic filter (equality)
                filters[prefixed_field_name] = graphql_type
                # browse & add available lookups
                for lookup_name, lookup_graphql_type in LOOKUPS.get(graphql_type, {}).items():
                    if not self.model_config.is_lookup_allowed(lookup_name):
                        continue
                    filters[f'{prefixed_field_name}__{lookup_name}'] = lookup_graphql_type
                    # apply same filters as for integers on date/time parts
                    date_time_types = (
                        graphql_types.Date, graphql_types.DateTime, graphql_types.Time)
                    if graphql_type in date_time_types and lookup_graphql_type == graphql_types.Int:
                        int_lookups = LOOKUPS[graphql_types.Int]
                        for int_lookup_name, int_lookup_graphql_type in int_lookups.items():
                            if not self.model_config.is_lookup_allowed(int_lookup_name):
                                continue
                            name = f'{prefixed_field_name}__{lookup_name}__{int_lookup_name}'
                            filters[name] = int_lookup_graphql_type
            elif isinstance(graphql_type, graphql.type.definition.GraphQLEnumType):
                filters[prefixed_field_name] = graphql_type
                if self.model_config.is_lookup_allowed('in'):
                    filters[f'{prefixed_field_name}__in'] = graphql_types.List(graphql_type)
                if self.model_config.is_lookup_allowed('isnull'):
                    filters[f'{prefixed_field_name}__isnull'] = graphql_types.Boolean
        # result
        return filters

    def get_table_name(self):
        raise NotImplementedError()

    # CRUD operations on ORM model instances

    def create_one(self, authenticated_user, graphql_path, graphql_selection, **data):
        """
            Create one instance of the given ORM model.

            Result is a `dict`, corresponding to the format given by `graphql_selection`.
        """
        raise NotImplementedError()

    def read_one(self, authenticated_user, graphql_path, graphql_selection, **filters):
        """
            Read one instance of the given ORM model.

            Result is a `dict`, corresponding to the format given by `graphql_selection`.
        """
        raise NotImplementedError()

    def read_many(self, authenticated_user, graphql_path, graphql_selection, **filters):
        """
            Read many instance of the given ORM model.

            Result is a `list` of `dict`, corresponding to the format given by `graphql_selection`.
        """
        raise NotImplementedError()

    def update_one(self, authenticated_user, graphql_path, graphql_selection, _=None, **filters):
        """
            Update one instance of the given ORM model.

            Result is a `dict`, corresponding to the format given by `graphql_selection`.
        """
        raise NotImplementedError()

    def delete_one(self, authenticated_user, graphql_path, graphql_selection, **filters):
        """
            Delete one instance of the given ORM model.

            Result is a `dict`, corresponding to the format given by `graphql_selection`.
        """
        raise NotImplementedError()

    #

    def _extract_custom_fields_data(self, operation, data):
        result = {}
        for custom_field in self.model_config.custom_fields:
            if custom_field.name in data and custom_field.can_perfom(operation):
                result[custom_field.name] = data.pop(custom_field.name)
        return result

    def _create_custom_fields(self, instance, authenticated_user, data):
        for custom_field in self.model_config.custom_fields:
            if custom_field.name in data:
                custom_field.perform_one_creation(
                    instance = instance,
                    authenticated_user = authenticated_user,
                    value = data[custom_field.name])

    def _update_custom_fields(self, instance, authenticated_user, data):
        for custom_field in self.model_config.custom_fields:
            if custom_field.name in data:
                custom_field.perform_one_update(
                    instance = instance,
                    authenticated_user = authenticated_user,
                    value = data[custom_field.name])

    def _read_custom_fields(self, instance, authenticated_user, graphql_selection):
        result = {}
        for custom_field in self.model_config.custom_fields:
            if custom_field.name in graphql_selection and custom_field.can_perfom(Operation.READ):
                result[custom_field.name] = custom_field.perform_one_read(
                    instance = instance,
                    authenticated_user = authenticated_user,
                    graphql_selection = graphql_selection)
        return result

    # methods should be executed within an atomic database transaction

    def decorate(self, method):
        """
            Decorator to execute a given method within a transaction, using
            the corresponding ORM.
        """
        raise NotImplementedError()

    # SQL logging

    @staticmethod
    def start_sql_log():
        """
            Clear ORM-level SQL log
        """
        raise NotImplementedError()

    @staticmethod
    def get_sql_log():
        """
            Get a log of SQL queries executed by the ORM in the form of a `list[str]`
        """
        raise NotImplementedError()



"""
    Definition of `DjangoModelManager` class.
"""

import django.db
import django.db.models
import django.db.transaction
from django.conf import settings
try:
    import django.contrib.postgres.fields
    WITH_POSTGRES_SUPPORT = True
except ImportError:
    WITH_POSTGRES_SUPPORT = False
import django.core.exceptions

from .. import graphql_types
from .. import conversion
from .. import exceptions
from ..operations import Operation
from .django_errors import reraise_from_django_validation_error
from ._manager import ModelManager
from ._fields import FieldsInfo, ForeignField, RelatedField


class DjangoModelManager(ModelManager):
    """
        ModelManager class for Django ORM.
    """

    # metadata extraction

    def get_fields_info(self):
        """
            Retrieve fields info for a given Django model.
        """
        # initialize result
        fields_info = FieldsInfo()
        # primary key
        fields_info.primary = self.orm_model._meta.pk.name # pylint: disable=protected-access
        # value & foreign fields_info
        for field in self.orm_model._meta.fields: # pylint: disable=protected-access
            # is it mandatory?
            if not field.blank:
                fields_info.mandatory |= {field.name, field.attname}
            # is it nullable?
            if field.null:
                fields_info.nullable |= {field.name, field.attname}
            # is it a foreign key?
            if isinstance(field, django.db.models.fields.related.ForeignKey):
                # field to which the foreign key is referring
                related_field = field.foreign_related_fields[0]
                # value field
                fields_info.value[field.attname] = self._to_graphql_type_from_field(related_field)
                # foreign field
                fields_info.foreign[field.name] = ForeignField(
                    orm_model = related_field.model,
                    field_name = field.name,
                    value_field_name = field.attname)
            # if not, it is a regular value field
            else:
                # value field
                fields_info.value[field.name] = self._to_graphql_type_from_field(field)
            # can it serve as a unique identifier?
            if field.unique:
                fields_info.unique[field.attname] = fields_info.value[field.attname]
        # related fields_info
        for related in self.orm_model._meta.related_objects: # pylint: disable=protected-access
            fields_info.related[related.name] = RelatedField(
                orm_model = related.related_model,
                field_name = related.field.name,
                value_field_name = related.field.attname)
        # return result
        return fields_info

    def get_table_name(self):
        return self.orm_model._meta.db_table

    # CRUD operations on ORM model instances

    def create_one(self, authenticated_user, graphql_path, graphql_selection=None, ensure_permission=True,
            depth=0, **data):
        # related things
        related_data = {}
        for field_name in list(data.keys()):
            if field_name in self.fields_info.foreign:
                foreign_field = self.fields_info.foreign[field_name]
                foreign_model_config = self.model_config.schema.get_model_config(
                    orm_model = foreign_field.orm_model)
                foreign_instance = foreign_model_config.orm_model_manager.create_one(
                    authenticated_user = authenticated_user,
                    graphql_path = graphql_path + [field_name],
                    ensure_permission = False,
                    depth = depth + 1,
                    **data.pop(field_name))
                data[field_name] = foreign_instance
            elif field_name in self.fields_info.related:
                related_data[field_name] = data.pop(field_name)
        # extract data for custom fields
        custom_fields_data = self._extract_custom_fields_data(
            operation = Operation.CREATE,
            data = data)
        # instance itself
        instance = self.orm_model(**data)
        # custom fields definition
        self._create_custom_fields(
            instance = instance,
            authenticated_user = authenticated_user,
            data = custom_fields_data)
        # pre-save trigger
        self.model_config.on_before_operation(instance, authenticated_user, Operation.CREATE, data, depth)
        # enforce permissions
        if ensure_permission:
            self.model_config.ensure_permission(
                operation = Operation.CREATE,
                instance = instance,
                authenticated_user = authenticated_user,
                graphql_path = graphql_path,
                data = data,
            )
        try:
            # validation
            instance.full_clean()
            # save
            instance.save()
        except django.core.exceptions.ValidationError as exception:
            reraise_from_django_validation_error(
                graphql_path, exception)
        # related data
        for field_name, children_data in related_data.items():
            for related_index, related_data in enumerate(children_data):
                related_field = self.fields_info.related[field_name]
                related_data[related_field.value_field_name] = instance.pk
                related_model_config = self.model_config.schema.get_model_config(
                    orm_model = related_field.orm_model)
                related_instance = related_model_config.orm_model_manager.create_one(
                    authenticated_user = authenticated_user,
                    graphql_path = graphql_path + [field_name, related_index],
                    ensure_permission = False,
                    depth = depth + 1,
                    **related_data)
                getattr(instance, field_name).add(related_instance)
        # validation
        try:
            if callable(getattr(instance, 'clean_related', None)):
                instance.clean_related()
        except django.core.exceptions.ValidationError as exception:
            reraise_from_django_validation_error(
                graphql_path, exception)
        # post-save trigger
        self.model_config.on_after_operation(instance, authenticated_user, Operation.CREATE, data, depth)
        # result
        if graphql_selection is None:
            return instance
        return self._instance_to_dict(
            authenticated_user = authenticated_user,
            instance = instance,
            graphql_selection = graphql_selection,
            graphql_path = graphql_path,
            ensure_permission = False,
        )

    def read_one(self, authenticated_user, graphql_path, graphql_selection, **filters):
        instance = self._read_one(
            graphql_selection = graphql_selection,
            authenticated_user = authenticated_user,
            **filters
        )
        return self._instance_to_dict(
            authenticated_user = authenticated_user,
            instance = instance,
            graphql_selection = graphql_selection,
            graphql_path = graphql_path,
            ensure_permission = True,
        )

    def read_many(self, authenticated_user, graphql_path, graphql_selection, **filters):
        # build queryset
        queryset = self._read(
            graphql_selection = graphql_selection,
            authenticated_user = authenticated_user,
            **filters
        )
        # build results, according to current limits
        if self.model_config.limit > -1:
            results = queryset[:self.model_config.limit]
        else:
            results = queryset.all()
        # return formatted result
        return [
            self._instance_to_dict(
                authenticated_user = authenticated_user,
                instance = instance,
                graphql_selection = graphql_selection,
                graphql_path = graphql_path,
                ensure_permission = False,
            )
            for instance in results
            if self.model_config.has_permission(
                operation = Operation.READ,
                instance = instance,
                authenticated_user = authenticated_user,
            )
        ]

    def update_one(self, authenticated_user, graphql_path, graphql_selection=None,
            _=None, depth=0, **filters):
        # variable that contains new data
        data = _ or {}
        # retrieve the instance to update
        instance = self._read_one(
            graphql_selection = graphql_selection or {},
            authenticated_user = authenticated_user,
            **filters
        )
        # enforce permissions
        self.model_config.ensure_permission(
            operation = Operation.UPDATE,
            instance = instance,
            authenticated_user = authenticated_user,
            graphql_path = graphql_path,
            data = data,
        )
        # pre-update trigger
        self.model_config.on_before_operation(instance, authenticated_user, Operation.UPDATE, data, depth)
        # related things
        related_data = {}
        for field_name in list(data.keys()):
            # foreign fields
            if field_name in self.fields_info.foreign:
                child_data = data.pop(field_name)
                # if child_data is null, the reference will be deleted
                if child_data is not None:
                    child_model_config = self.model_config.schema.get_model_config(
                        orm_model = self.fields_info.foreign[field_name].orm_model)
                    child_identifier = child_data.pop(
                        child_model_config.orm_model_manager.fields_info.primary, None)
                    # if no identifier provided, create a new instance
                    if child_identifier is None:
                        data[field_name] = child_model_config.orm_model_manager.create_one(
                            authenticated_user = authenticated_user,
                            graphql_path = graphql_path + [field_name],
                            depth = depth + 1,
                            **child_data)
                    # if identifier provided, update existing instance
                    else:
                        data[field_name] = child_model_config.orm_model_manager.update_one(
                            authenticated_user = authenticated_user,
                            graphql_path = graphql_path + [field_name],
                            _ = child_data,
                            depth = depth + 1,
                            **{child_model_config.orm_model_manager.fields_info.primary:
                                child_identifier})
            # related fields
            elif field_name in self.fields_info.related:
                related_data[field_name] = data.pop(field_name)
        # extract data for custom fields
        custom_fields_data = self._extract_custom_fields_data(
            operation = Operation.UPDATE,
            data = data)
        # direct attributes
        for key, value in data.items():
            setattr(instance, key, value)
        # custom fields definition
        self._update_custom_fields(
            instance = instance,
            authenticated_user = authenticated_user,
            data = custom_fields_data)
        # validate & save (raise an easy_graphql_server exception instead of a Django one)
        try:
            # validation
            instance.full_clean()
            # save instance
            instance.save()
        except django.core.exceptions.ValidationError as exception:
            reraise_from_django_validation_error(
                graphql_path, exception)
        # related data
        for field_name, children_data in related_data.items():
            related_field = self.fields_info.related[field_name]
            child_model_config = self.model_config.schema.get_model_config(
                orm_model = related_field.orm_model)
            children_identifiers = []
            # create & update
            for child_index, child_data in enumerate(children_data):
                child_identifier = child_data.get(
                    child_model_config.orm_model_manager.fields_info.primary, None)
                # update only if identifier provided
                if child_identifier is not None:
                    child_instance = child_model_config.orm_model_manager.update_one(
                        authenticated_user = authenticated_user,
                        graphql_path = graphql_path + [field_name, child_index],
                        _ = child_data,
                        depth = depth + 1,
                        **{child_model_config.orm_model_manager.fields_info.primary:
                            child_identifier})
                    # store identifier
                    children_identifiers.append(child_identifier)
            # delete omitted children
            for child_instance in getattr(instance, field_name).all():
                if child_instance.pk not in children_identifiers:
                    child_instance.delete()
            # create
            for child_index, child_data in enumerate(children_data):
                child_identifier = child_data.get(
                    child_model_config.orm_model_manager.fields_info.primary, None)
                # create only if no identifier provided
                if child_identifier is None:
                    child_instance = child_model_config.orm_model_manager.create_one(
                        authenticated_user = authenticated_user,
                        graphql_path = graphql_path + [field_name, child_index],
                        depth = depth + 1,
                        **dict({related_field.value_field_name: instance.pk}, **child_data))
                    getattr(instance, field_name).add(child_instance)
        # validation (raise an easy_graphql_server exception instead of a Django one)
        try:
            instance.full_clean()
            if callable(getattr(instance, 'clean_related', None)):
                instance.clean_related()
        except django.core.exceptions.ValidationError as exception:
            reraise_from_django_validation_error(
                graphql_path, exception)
        # post-update trigger
        self.model_config.on_after_operation(instance, authenticated_user, Operation.UPDATE, data, depth)
        # result
        if graphql_selection is None:
            return instance
        return self._instance_to_dict(
            authenticated_user = authenticated_user,
            instance = instance,
            graphql_selection = graphql_selection,
            graphql_path = graphql_path,
            ensure_permission = True,
        )

    def delete_one(self, authenticated_user, graphql_path, graphql_selection, **filters):
        instance = self._read_one(
            graphql_selection = graphql_selection,
            authenticated_user = authenticated_user,
            **filters
        )
        self.model_config.ensure_permission(
            operation = Operation.DELETE,
            instance = instance,
            authenticated_user = authenticated_user,
            graphql_path = graphql_path,
        )
        # keep primary key for later result
        if self.fields_info.primary in graphql_selection:
            pre_result = {self.fields_info.primary: getattr(instance, self.fields_info.primary)}
        else:
            pre_result = {}
        # actually perform operations
        self.model_config.on_before_operation(instance, authenticated_user, Operation.DELETE, None, 0)
        instance.delete()
        self.model_config.on_after_operation(instance, authenticated_user, Operation.DELETE, None, 0)
        # compute & return result
        result = self._instance_to_dict(
            authenticated_user = authenticated_user,
            instance = instance,
            graphql_selection = graphql_selection,
            graphql_path = graphql_path,
            ensure_permission = False,
        )
        result.update(pre_result)
        return result

    # methods should be executed within an atomic database transaction

    def decorate(self, method, graphql_path=None):
        """
            Every exposed method will have to go through this decorator.
        """
        def decorated(*args, **kwargs):
            with django.db.transaction.atomic():
                try:
                    return method(*args, **kwargs)
                except django.core.exceptions.ValidationError as exception:
                    reraise_from_django_validation_error(
                        graphql_path or kwargs['graphql_path'], exception)
        return decorated

    # helpers for reading

    def _read(self, graphql_selection, authenticated_user, **filters):
        # build queryset as intended by easy_graphql_server
        queryset = self.build_queryset(
            graphql_selection = graphql_selection,
            authenticated_user = authenticated_user
        ).filter(**filters)
        # filter queryset, depending on model config
        queryset = self.model_config.filter_for_user(
            queryset = queryset,
            authenticated_user = authenticated_user)
        # return resulting queryset
        return queryset

    def _read_one(self, graphql_selection, authenticated_user, **filters):
        try:
            return self._read(
                graphql_selection = graphql_selection,
                authenticated_user = authenticated_user,
                **filters
            ).get()
        except django.core.exceptions.ObjectDoesNotExist as error:
            raise exceptions.NotFoundError(filters) from error

    def _instance_to_dict(self, instance, authenticated_user, graphql_selection, graphql_path,
            ensure_permission=True, depth=0):
        # pre-read trigger
        self.model_config.on_before_operation(instance, authenticated_user, Operation.READ, None, depth)
        # enforce permissions when requested
        if ensure_permission:
            self.model_config.ensure_permission(
                operation = Operation.READ,
                instance = instance,
                authenticated_user = authenticated_user,
                graphql_path = graphql_path,
            )
        # build result: custom fields
        result = self._read_custom_fields(
            instance = instance,
            authenticated_user = authenticated_user,
            graphql_selection = graphql_selection)
        # build result: from instance attributes
        for field_name, graphql_subselection in graphql_selection.items():
            if field_name in result:
                continue
            field_value = getattr(instance, field_name)
            # field_value field
            if graphql_subselection is None:
                result[field_name] = field_value
            # related field
            elif type(field_value).__name__ == 'RelatedManager':
                children_instances = field_value.all()
                if len(children_instances) == 0:
                    result[field_name] = []
                else:
                    model_manager = self.model_config.schema.get_model_config(
                        orm_model=children_instances[0].__class__).orm_model_manager
                    result[field_name] = [
                        model_manager._instance_to_dict(
                            authenticated_user = authenticated_user,
                            instance = child_instance,
                            graphql_selection = graphql_subselection,
                            graphql_path = graphql_path + [field_name, child_index],
                            ensure_permission = False,
                            depth = depth + 1,
                        )
                        for child_index, child_instance
                        in enumerate(children_instances)
                    ]
            # foreign field
            elif field_value is not None:
                model_manager = self.model_config.schema.get_model_config(
                    orm_model=field_value.__class__).orm_model_manager
                # pylint: disable=protected-access
                result[field_name] = model_manager._instance_to_dict(
                    authenticated_user = authenticated_user,
                    instance = field_value,
                    graphql_selection = graphql_subselection,
                    graphql_path = graphql_path + [field_name],
                    ensure_permission = ensure_permission,
                    depth = depth + 1,
                )
        # post-read trigger
        self.model_config.on_after_operation(instance, authenticated_user, Operation.READ, result, depth)
        return result

    def build_queryset(self, graphql_selection, authenticated_user):
        """
            Build queryset for given GraphQL selection
        """
        only, prefetch_related, select_related = (
            self.build_queryset_parts(
                graphql_selection = graphql_selection,
                authenticated_user = authenticated_user,
            )
        )
        base_queryset = self.model_config.filter_for_user(self.orm_model.objects, authenticated_user)
        if self.restrict_queried_fields:
            return (base_queryset
                .only(*only)
                .prefetch_related(*prefetch_related)
                .select_related(*select_related)
            )
        return (base_queryset
            .prefetch_related(*prefetch_related)
            .select_related(*select_related)
        )

    def build_queryset_parts(self, graphql_selection, authenticated_user,
            field_prefix=''):
        """
            Build queryset parts for given GraphQL selection

            Parts are returned as a tuple of these four values:
             - the base queryset
             - a list of the only fields to select
             - a list of what should be passed to `QuerySet.prefetch_related()`
             - a list of what should be passed to `QuerySet.select_related()`
        """
        schema = self.model_config.schema
        # initialize result
        only = []
        prefetch_related = []
        select_related = []
        # browse fields in GraphQL selection
        for field_name, graphql_subselection in graphql_selection.items():
            # no subselection, this is a direct field
            if graphql_subselection is None and field_name not in self.fields_info.custom:
                only.append(field_prefix + field_name)
            # there is a subselection, and it is a foreign key
            elif field_name in self.fields_info.foreign:
                foreign_field = self.fields_info.foreign[field_name]
                only.append(field_prefix + foreign_field.value_field_name)
                select_related.append(field_prefix + field_name)
                foreign_orm_model_manager = schema.get_model_config(
                    orm_model = foreign_field.orm_model).orm_model_manager
                foreign_only, foreign_prefetch_related, foreign_select_related = (
                    foreign_orm_model_manager.build_queryset_parts(
                        graphql_selection = graphql_subselection,
                        authenticated_user = authenticated_user,
                        field_prefix = f'{field_prefix}{field_name}__'
                    )
                )
                only += foreign_only
                prefetch_related += foreign_prefetch_related
                select_related += foreign_select_related
            # there is a subselection, and it is a related field
            elif field_name in self.fields_info.related:
                related_field = self.fields_info.related[field_name]
                related_model_config = schema.get_model_config(
                    orm_model = related_field.orm_model)
                prefetch_related.append(
                    django.db.models.Prefetch(
                        f'{field_prefix}{field_name}',
                        queryset = related_model_config.orm_model_manager.build_queryset(
                            authenticated_user = authenticated_user,
                            graphql_selection = dict(
                                {related_field.value_field_name: None},
                                **graphql_subselection),
                        )
                    )
                )
        # return resulting queryset
        return only, prefetch_related, select_related


    # Django field conversion to GraphQL type

    GRAPHQL_TYPES_MAPPING = {
        # boolean
        django.db.models.fields.BooleanField: graphql_types.Boolean,
        # integers
        django.db.models.fields.AutoField: graphql_types.Int,
        django.db.models.fields.IntegerField: graphql_types.Int,
        django.db.models.fields.BigIntegerField: graphql_types.Int,
        django.db.models.fields.SmallIntegerField: graphql_types.Int,
        # non-integer numbers
        django.db.models.fields.FloatField: graphql_types.Float,
        django.db.models.fields.DecimalField: graphql_types.Decimal,
        # texts
        django.db.models.fields.CharField: graphql_types.String,
        django.db.models.fields.TextField: graphql_types.String,
        django.db.models.fields.URLField: graphql_types.String,
        django.db.models.fields.EmailField: graphql_types.String,
        # date/time
        django.db.models.fields.DateField: graphql_types.Date,
        django.db.models.fields.DateTimeField: graphql_types.DateTime,
        django.db.models.fields.TimeField: graphql_types.Time,
        # other things
        django.db.models.fields.json.JSONField: graphql_types.JSONString,
    }

    def _to_graphql_type_from_field(self, field):
        """
            Convert Django ORM model fields to GraphQL type.
        """
        # enum
        if getattr(field, 'choices', None):
            graphql_type = conversion.to_graphql_enum_from_choices(
                prefix = f'{self.model_config.types_name}__{field.name}',
                choices = field.choices,
                schema = self.model_config.schema,
            )
        # scalar
        elif isinstance(field, tuple(self.GRAPHQL_TYPES_MAPPING)):
            graphql_type = self.GRAPHQL_TYPES_MAPPING[type(field)]
        # Postgres
        elif WITH_POSTGRES_SUPPORT:
            # array
            if isinstance(field, django.contrib.postgres.fields.array.ArrayField):
                graphql_type = graphql_types.List(
                    self._to_graphql_type_from_field(field.base_field))
        # unrecognized
        else:
            raise ValueError(f'Could not conversion {field} to graphql type')
        # result
        return graphql_type

    # SQL logging

    @staticmethod
    def start_sql_log():
        settings.DEBUG = True
        django.db.reset_queries()

    @staticmethod
    def get_sql_log():
        return [
            query['sql']
            for query in list(django.db.connection.queries)
        ]



"""
    This module defines `Required` class and `JSONString` type.
"""

from .graphql_types import JSONString # pylint: disable=unused-import
from .operations import Operation


class Required:
    # pylint: disable=too-few-public-methods
    """
        Non-GraphQL wrapper type, to replace NonNull in mappings or when using
        "natural" Python types
    """
    def __init__(self, type_):
        self.type_ = type_


class ModelField:
    """
        Non-GraphQL wrapper type, to use the same type as a field of an already exposed model.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, model_name, field_path=None):
        self.model_name = model_name
        self.field_path = field_path or []

    def __getattr__(self, field_name):
        return self.__class__(
            model_name = self.model_name,
            field_path = self.field_path + [field_name],
        )

    __getitem__ = __getattr__

class ModelInterface:
    """
        Non-GraphQL wrapper type, to describe the interface of an already exposed model.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, model_name, operation, exclude=None, additional=None):
        self.model_name = model_name
        self.operation = operation
        self.exclude = exclude or set()
        self.additional = additional or {}

    def __add__(self, fields):
        return self.__class__(
            model_name = self.model_name,
            operation = self.operation,
            exclude = self.exclude,
            additional = dict(self.additional, **fields))

    def __sub__(self, fields):
        return self.__class__(
            model_name = self.model_name,
            operation = self.operation,
            exclude = self.exclude | set(fields),
            additional = self.additional)

class Model:
    """
        Non-GraphQL wrapper type, to use the same interface as an already exposed model.
    """

    def __init__(self, model_name):
        self.model_name = model_name

    @property
    def create_input_format(self):
        """
            To use the same type as the input of the create mutation
        """
        return ModelInterface(self.model_name, Operation.CREATE)

    @property
    def update_input_format(self):
        """
            To use the same type as the input of the update mutation
        """
        return ModelInterface(self.model_name, Operation.UPDATE)

    @property
    def output_format(self):
        """
            To use the same type as the output format of the exposed model
        """
        return ModelInterface(self.model_name, Operation.READ)

    @property
    def fields(self):
        """
            A mapping to all the fields of the model, regardless of what is exposed
            for this model
        """
        return ModelField(self.model_name)



"""
    Here are defined some managers to help dealing with case.
"""


import re
import enum


class CaseManager:
    """
        Base case manager, inherited by `PascalCaseManager` and `SnakeCaseManager`
    """

    @classmethod
    def convert_and_join(cls, *names):
        """
            Convert names, and then join them together using the corresponding convention.
        """
        return cls.join(
            cls.convert(name)
            for name in names
        )

    @staticmethod
    def convert(name):
        """
            This method should be overriden in classes inheriting from `CaseManager`.
        """
        raise NotImplementedError()

    @staticmethod
    def join(name):
        """
            This method should be overriden in classes inheriting from `CaseManager`.
        """
        raise NotImplementedError()


class PascalCaseManager(CaseManager):
    """
        Case manager for Pascal casing convention.
    """

    @staticmethod
    def convert(name):
        """
            Convert a name to Pascal case.
        """
        return re.sub(
            r'(?:^|_)([a-zA-Z])([^_]+)',
            lambda m: m.group(1).upper() + m.group(2),
            name
        )

    @staticmethod
    def join(*names):
        """
            Join names using Pascal case convention.
        """
        return ''.join(
            name[0].upper() + name[1:]
            for name in names
        )


class SnakeCaseManager(CaseManager):
    """
        Case manager for snake casing convention.
    """

    @staticmethod
    def convert(name):
        """
            Convert a name to snake case.
        """
        return re.sub(
            r'([a-z])([A-Z])',
            lambda m: f'{m.group(1)}_{m.group(2)}',
            name
        ).lower()

    @staticmethod
    def join(*names):
        """
            Join names using snake case convention.
        """
        return '_'.join(names)


@enum.unique
class Casing(enum.Enum):
    """
        This enum lists available casing styles, and also provides with
        corresponding case managers.
    """
    PASCAL = PascalCaseManager
    SNAKE = SnakeCaseManager



#!/usr/bin/env python

from re import search
from setuptools import setup, find_packages

version = '0.9.26',

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='easy_graphql_server',
    version = '0.9.26',
    description='Easy to use abstraction layer for GraphQL, with support for Django ORM.',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords='graphql django',
    url='https://github.com/mathieurodic/easy-graphql-server',
    author='Mathieu Rodic',
    author_email='graphql@rodic.fr',
    license='MIT license',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires=[
        'graphql-core==3.1.7'
    ],
    extras_require={
        'tests': [
            'django==3.2',
            'faker==11.1.0',
            'flask==2.0.2',
            'flask-login==0.5.0',
        ],
    },
    python_requires='>=3.6,<4',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    # PEP-561: https://www.python.org/dev/peps/pep-0561/
    package_data={
        'easy_graphql_server': ['py.typed'],
        'easy_graphql_server.webserver.static': ['*'],
    },
    include_package_data=True,
    zip_safe=False,
)



"""
    Data-driven tests for schema defined with `easy_graphql_server.Schema`.
"""

import os
import re
import json
import importlib.util
import unittest


DefaultBaseTestCase = unittest.TestCase


def generate_testcase(schema, graphql_path,
        base_test_class=DefaultBaseTestCase, django_environment=False):
    """
        Generate one test case class, corresponding to a `.gql` file.
    """
    # compute test class name
    class_name = graphql_path.split(os.sep)[-1].split('.')[0]
    class_name = re.sub(r'[^\w]+', '_', class_name).lower()
    class_name = f'easy_graphql_server_test_{class_name}'
    # generate test case from files
    class TestCase(base_test_class):
        """
            Base test case class, associated with a `.gql` file.
        """
        @staticmethod
        def _replace_extension(path, new_extension):
            return re.sub(r'\.\w+$', f'.{new_extension}', path)
        def _iterate_data(self, graphql_path):
            # load GraphQL input
            with open(graphql_path, 'rt', encoding='utf-8') as graphql_file:
                graphql_list = re.split(r'\n\s*;\s*\n', graphql_file.read())
            # load JSON output
            json_path = self._replace_extension(graphql_path, 'json')
            with open(json_path, 'rt', encoding='utf-8') as json_file:
                json_data = json_file.read()
                json_data = re.sub(r'(^|\n)//[^\n]+', r'\1', json_data)
                json_list = []
                for i, expected_output in enumerate(re.split(r'\n\s*;\s*\n', json_data)):
                    try:
                        json_list.append(
                            json.loads(expected_output)
                        )
                    except json.decoder.JSONDecodeError as error:
                        raise ValueError(f'Cannot decode JSON in `{json_path}`[{i}]: '
                            '{error}\n{expected_output}') from error
            # load generated SQL
            sql_path = self._replace_extension(graphql_path, 'sql')
            if os.path.isfile(sql_path):
                with open(sql_path, 'rt', encoding='utf-8') as sql_file:
                    sql_list = sql_file.read().split(';;')
            else:
                sql_list = []
            sql_list += (len(graphql_list) - len(sql_list)) * ['']
            # return iterator
            for i, (graphql, json_, sql) in enumerate(zip(graphql_list, json_list, sql_list)):
                # handle USER directive
                username_match = re.search(r'(?:^|\n)\s*#\s*USER\s*:\s*([^\s]+)', graphql)
                username = username_match.group(1) if username_match else None
                if username is None:
                    user = None
                elif not hasattr(self, 'get_or_create_user'):
                    raise AttributeError(
                        'To use the `USER` directive, you must implement a `get_or_create_user(username) '
                        'method` on a base class for test cases, and pass this class as a '
                        '`base_test_class` parameter.')
                else:
                    user = self.get_or_create_user(username)
                # yield expected data
                yield i, (user, graphql, json_, sql.strip())
        @staticmethod
        def show_diff(path, index, expectation, reality, input_=None):
            """
                Show difference between expected and actual result.
            """
            print(90 * '~')
            print(f'{path} : {index}')
            if input_:
                print(40 * '~' + ' INPUT IS ' + 40 * '~')
                print(input_)
            print(40 * '-' + ' EXPECTED ' + 40 * '-')
            print(expectation)
            print(40 * '+' + ' RECEIVED ' + 40 * '+')
            print(reality)
            print(90 * '~')
        def run_test(self, *args, **kwargs): # pylint: disable=unused-argument
            """
                Run data-driven test
            """
            # show the difference, no matter how long
            self.maxDiff = None # pylint: disable=invalid-name
            # so we can debug SQL queries
            if django_environment:
                from django.conf import settings # pylint: disable=import-outside-toplevel
                settings.DEBUG = True
            # execute setup script if present
            setup_script_path = self._replace_extension(graphql_path, 'setup.py')
            if os.path.isfile(setup_script_path):
                spec = importlib.util.spec_from_file_location(
                    f'setup_{class_name}', setup_script_path)
                setup_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(setup_module)
            # actually run the test step(s)
            for index, (user, graphql, json_, sql) in self._iterate_data(graphql_path):
                # clear sql log
                if sql:
                    schema.check()
                    for orm_model_manager_class in set(schema.orm_model_manager_classes):
                        orm_model_manager_class.start_sql_log()
                # compute output
                json_result = schema.execute(
                    query = graphql,
                    authenticated_user = user,
                    serializable_output = True,
                )
                try:
                    # compare it with what was expected
                    self.assertEqual(json_, json_result)
                except:
                    # if different, show pretty output and re-raise assert exception
                    self.show_diff(
                        path = self._replace_extension(graphql_path, 'json'),
                        index = index,
                        expectation = json.dumps(json_, indent=2),
                        reality = json.dumps(json_result, indent=2),
                        input_ = re.sub(r'(?:(?<=\n)|^)[ \t]*#(?![ \t]*USER)[^\n]*(?:\n|$)', '', graphql),
                    )
                    raise
                # check SQL also (when provided)
                if sql:
                    sql_result_list = []
                    for orm_model_manager_class in set(schema.orm_model_manager_classes):
                        sql_result_list += orm_model_manager_class.get_sql_log()
                    sql_result = '\n;\n'.join(sql_result_list)
                    try:
                        # compare it with what was expected
                        self.assertEqual(sql, sql_result)
                    except:
                        self.show_diff(
                            path = self._replace_extension(graphql_path, "sql"),
                            index = index,
                            expectation = sql,
                            reality = sql_result,
                        )
                        raise

    # set test case class name
    TestCase.__name__ = class_name
    TestCase.__qualname__ = '.'.join(TestCase.__qualname__.split('.')[:-1] + [class_name])
    # return generated test case class
    return TestCase('run_test')

def _get_gql_path_list(path):
    if os.path.isfile(path) and path.endswith('.gql'):
        yield path
    elif os.path.isdir(path):
        for entry in os.scandir(path):
            yield from _get_gql_path_list(entry.path)

def generate_testcases(schema, path, base_test_class=DefaultBaseTestCase):
    """
        Iterator to generate all test cases for a schema, given a path.

        `path` can either be a directory that will be recursively searched for `.gql` files,
        or a `.gql` file.

        Each of the `.gql` file will correspond to a yielded test case class, generated
        with `generate_testcase()`.
    """
    for gql_path in sorted(_get_gql_path_list(path)):
        yield generate_testcase(schema, gql_path, base_test_class)

def make_tests_loader(schemata, path, base_test_class=DefaultBaseTestCase):
    """
        This method returns a `load_tests` method, useful for loading generated tests
        from a module.

        Parameters have the same meaning as for `generate_testcases()`.

        Example:

        ```python
        from easy_graphql_server import Schema
        from easy_graphql_server.testing import make_tests_loader

        schema = Schema()

        load_tests = make_tests_loader(schema, 'path/to/graphql/test/data')
        ```
    """
    if not isinstance(schemata, (tuple, list, set)):
        schemata = [schemata]
    def load_tests(loader, tests, ignore): # pylint: disable=unused-argument
        path_ = os.getenv('EASY_GRAPHQL_SERVER_TESTS_PATH', path)
        for schema in schemata:
            for test in generate_testcases(schema, path_, base_test_class):
                tests.addTest(test)
        return tests
    return load_tests



"""
    This module defines the `ModelConfigCustomField` class.
"""

from .operations import Operation

class ModelConfigCustomField:
    """
        Configuration of a custom field added to an exposed model.
    """

    def __init__(self, name, format, read_one=None, read_many=None,
            update_one=None, update_many=None, create_one=None, create_many=None):
        self.name = name
        self.format = format
        self.read_one = read_one
        self.read_many = read_many
        self.update_one = update_one
        self.update_many = update_many
        self.create_one = create_one
        self.create_many = create_many

    def can_perfom(self, operation):
        """
            Check if the given operation can be performed on that field
        """
        if operation == Operation.CREATE:
            return self.create_one or self.create_many
        if operation == Operation.READ:
            return self.read_one or self.read_many
        if operation == Operation.UPDATE:
            return self.update_one or self.update_many
        raise NotImplementedError()

    def perform_one_read(self, instance, authenticated_user, graphql_selection):
        """
            Fetch one value for the custom field for one instance
        """
        if self.read_one:
            return self.read_one(
                instance=instance,
                authenticated_user=authenticated_user,
                graphql_selection=graphql_selection)
        if self.read_many:
            return list(self.read_many(
                instances=[instance],
                authenticated_user=authenticated_user,
                graphql_selection=graphql_selection))[0]
        raise NotImplementedError()

    def perform_many_reads(self, instances, authenticated_user, graphql_selection):
        """
            Fetch many values for the custom field of many instance
        """
        if self.read_many:
            return self.read_many(
                instances=instances,
                authenticated_user=authenticated_user,
                graphql_selection=graphql_selection)
        if self.read_one:
            return [
                self.read_one(
                    instance=instance,
                    authenticated_user=authenticated_user,
                    graphql_selection=graphql_selection)
                for instance in instances
            ]
        raise NotImplementedError()

    def perform_one_creation(self, instance, authenticated_user, value):
        """
            Update the custom field for one instance
        """
        if self.create_one:
            return self.create_one(
                instance=instance,
                authenticated_user=authenticated_user,
                value=value)
        if self.create_many:
            return list(self.create_many(
                instances=[instance],
                authenticated_user=authenticated_user,
                value=value))[0]
        raise NotImplementedError()

    def perform_one_update(self, instance, authenticated_user, value):
        """
            Update the custom field for one instance
        """
        if self.update_one:
            return self.update_one(
                instance=instance,
                authenticated_user=authenticated_user,
                value=value)
        if self.update_many:
            return list(self.update_many(
                instances=[instance],
                authenticated_user=authenticated_user,
                value=value))[0]
        raise NotImplementedError()



"""
    Definition of `FlaskSchemaView` class
"""

from flask import request
import flask_login

from ._schema_view import SchemaView


class FlaskSchemaView(SchemaView):

    """
        Flask schema view. Used when calling `Schema.as_flask_view()`.
    """

    def view(self):
        """
            Flask view to compute GraphQL request
        """
        return self.compute_response(
            method = request.method,
            headers = request.headers,
            body = request.get_data(as_text=True),
            query = request.args.to_dict(),
            authenticated_user = flask_login.current_user,
        )



import inspect

from .. import exceptions


def reraise_from_django_validation_error(graphql_path, exception):

    def serialize(issue, path, field_name=None):
        if hasattr(issue, 'error_dict'):
            for field, errors in issue.error_dict.items():
                for error in errors:
                    yield from serialize(error, path, field)
        else:
            for error in issue.error_list:
                # extract & format params
                params = getattr(error, 'params', None) or {}
                for key in list(params.keys()):
                    if isinstance(params[key], (str, int, float, list, dict)):
                        continue
                    if isinstance(params[key], (tuple, set)):
                        params[key] = list(params[key])
                    elif inspect.isclass(params[key]):
                        params[key] = params[key].__name__
                    else:
                        params[key] = str(params[key])
                # extract & format message
                message = error.message
                if message and error.params:
                    message %= params
                # yield one serialized error
                yield {
                    'path': path + [field_name] if field_name else path,
                    'message': str(message),
                    'params': getattr(error, 'params', {}),
                    'code': getattr(error, 'code', None),
                }
    issues = list(serialize(exception, graphql_path))
    raise exceptions.ValidationError(issues) from exception
